"""payu.models.staged_cable
   ================

   Driver interface to CABLE-POP_TRENDY branch

   :copyright: Copyright 2021 Marshall Ward, see AUTHORS for details
   :license: Apache License, Version 2.0, see LICENSE for details
"""

# Standard Library
import os
import shutil
import itertools

# Extensions
import f90nml
import yaml

# Local
from payu.models.model import Model
from payu.fsops import mkdir_p


def deep_update(d_1, d_2):
    """Deep update of namelists."""
    for key, value in d_2.items():
        if isinstance(value, dict):
            # Nested struct
            if key in d_1:
                # If the master namelist contains the key, then recursively
                # apply
                deep_update(d_1[key], d_2[key])
            else:
                # Otherwise just set the value from the patch dict
                d_1[key] = value
        else:
            # Is value, just override
            d_1[key] = value


class StagedCable(Model):
    """A driver for running staged CABLE spin-up configurations."""

    def __init__(self, expt, name, config):
        super(StagedCable, self).__init__(expt, name, config)

        self.model_type = 'staged_cable'
        self.default_exec = 'cable'

        self.config_files = ['stage_config.yaml']
        self.optional_config_files = ['cable.nml', 'cru.nml',
                                      'luc.nml', 'met_names.nml']

        self.configuration_log = {}

        if not os.path.isfile('configuration_log.yaml'):
            # Build a new configuration log
            self._build_new_configuration_log()
        else:
            # Read the current configuration log
            self._read_configuration_log()

        # Now set the number of runs using the configuration_log
        remaining_stages = len(self.configuration_log['queued_stages'])
        print("Overriding the remaining number of runs according to the " +
              "number of queued stages in the configuration log.")
        os.environ['PAYU_N_RUNS'] = str(remaining_stages)

    def _build_new_configuration_log(self):
        """Build a new configuration log for the first stage of the run."""

        # Read the stage_config.yaml file
        with open('stage_config.yaml', 'r') as stage_conf_f:
            self.stage_config = yaml.safe_load(stage_conf_f)

        # On the first run, we need to read the 'stage_config.yaml' file.
        cable_stages = self._prepare_configuration()

        # Build the configuration log
        self.configuration_log['queued_stages'] = cable_stages
        self.configuration_log['current_stage'] = ''
        self.configuration_log['completed_stages'] = []

        self._save_configuration_log()

    def _read_configuration_log(self):
        """Read the existing configuration log."""
        with open('configuration_log.yaml') as conf_log_file:
            self.configuration_log = yaml.safe_load(conf_log_file)

    def _prepare_configuration(self):
        """Prepare the stages in the CABLE configuration."""

        # Since Python3.7, dictionary order is guaranteed so we can read
        # the entries in order without needing to supply an index
        # We just want to populate cable_stages with the list of stages
        # to run
        cable_stages = []
        for stage_name, stage_opts in self.stage_config.items():
            # Check if stage is a multi-step or single step
            if stage_name.startswith('multistep'):
                # The multi-step stage can run each internal stage
                # a different number of times. For example, a two
                # step stage may ask for the first step (S1) 5 times,
                # but the second step (S2) only 3 times. The stage
                # looks like [S1, S2, S1, S2, S1, S2, S1, S1].
                # So what we need to do is first record the number
                # of times each step is run

                # Use recipe from https://stackoverflow.com/questions/48199961
                # Turn the stages into lists of "count" length
                steps = [[step_name] * stage_opts[step_name]['count']
                         for step_name in stage_opts.keys()]

                cable_stages.extend(
                    [stage for stage in itertools.chain.from_iterable(
                        itertools.zip_longest(*steps)
                    ) if stage is not None]
                )
                # Finish handling of multistep stage

            else:
                # A single step stage, in general we only want to run this
                # once, but check for the count anyway
                cable_stages.extend([stage_name] * stage_opts['count'])

                # Finish handling of single step stage
        return cable_stages

    def setup(self):
        super(StagedCable, self).setup()

        # Prepare the namelists for the stage
        stage_name = self._get_stage_name()
        self._apply_stage_namelists(stage_name)

        # Make the logging directory
        mkdir_p(os.path.join(self.work_path, "logs"))

        # Get the additional restarts from older restart dirs
        self._get_further_restarts()

        # Make necessary adjustments to the configuration log
        self._handle_configuration_log_setup()

    def _get_further_restarts(self):
        """Get the restarts from stages further in the past where necessary."""

        # Often we take restarts from runs which are not the most recent run as
        # inputs for particular science modules, which means we have to extend
        # the existing functionality around retrieving restarts.

        # We can't supercede the parent get_prior_restart_files, since the
        # files returned by said function are prepended by
        # self.prior_restart_path, which is not desirable in this instance.

        num_completed_stages = len(self.configuration_log['completed_stages'])

        for stage_number in reversed(range(num_completed_stages - 1)):
            respath = os.path.join(
                self.expt.archive_path,
                f'restart{stage_number:03d}'
            )
            for f_name in os.listdir(respath):
                if os.path.isfile(os.path.join(respath, f_name)):
                    f_orig = os.path.join(respath, f_name)
                    f_link = os.path.join(self.work_init_path_local, f_name)
                    # Check whether a given link already exists in the
                    # manifest, so we don't write over a newer version of a
                    # restart
                    if f_link not in self.expt.manifest.manifests['restart']:
                        self.expt.manifest.add_filepath(
                            'restart',
                            f_link,
                            f_orig,
                            self.copy_restarts
                        )

    def set_model_pathnames(self):
        super(StagedCable, self).set_model_pathnames()

        self.work_restart_path = os.path.join(self.work_path, 'restart')
        self.work_output_path = os.path.join(self.work_path, 'outputs')

    def _get_stage_name(self):
        """Get the name of the stage being prepared."""

        if self.configuration_log['current_stage'] != '':
            # If the current stage is a non-empty string, it means we exited
            # during the running of the previous stage
            stage_name = self.configuration_log['current_stage']
        else:
            # Pop the stage from the list
            stage_name = self.configuration_log['queued_stages'][0]

        # Ensure the directory exists
        if not os.path.isdir(os.path.join(self.control_path, stage_name)):
            errmsg = f"""Directory containing namelists for stage {stage_name}
             does not exist."""
            raise FileNotFoundError(errmsg)

        return stage_name

    def _apply_stage_namelists(self, stage_name):
        """Apply the stage namelists to the master namelists.

        The master namelist is the namelist that exists in the control
        directory and the stage namelist exists within the directory for the
        given stage. If a master version of a given namelist does not exist,
        then the stage namelist is taken as is.

        Example:
        .
        ├── cable.nml
        └── cable_stage
            ├── cable.nml
            └── luc.nml

        In this instance, the ```cable.nml``` for ```cable_stage``` would be
        a merge of the top level ```cable.nml``` and
        ```cable_stage/cable.nml``` (with the latter taking precedence) and
        ```luc.nml``` is just ```cable_stage/luc.nml```.
        """
        namelists = os.listdir(os.path.join(self.control_path, stage_name))

        for namelist in namelists:
            write_target = os.path.join(self.work_input_path, namelist)
            stage_nml = os.path.join(self.control_path, stage_name, namelist)

            if os.path.isfile(os.path.join(self.control_path, namelist)):
                # Instance where there is a master and stage namelist
                with open(stage_nml) as stage_nml_f:
                    stage_namelist = f90nml.read(stage_nml_f)

                master_nml = os.path.join(self.control_path, namelist)
                f90nml.patch(master_nml, stage_namelist, write_target)
            else:
                # Instance where there is only a stage namelist
                shutil.copy(stage_nml, write_target)

    def _handle_configuration_log_setup(self):
        """Make appropriate adjustments to the configuration log to reflect
        that the setup of the stage is complete."""

        if self.configuration_log['current_stage'] != '':
            # If the current stage is a non-empty string, it means we exited
            # during the running of the previous stage- leave as is
            stage_name = self.configuration_log['current_stage']
        else:
            # Normal case where we just archived a successful stage.
            self.configuration_log['current_stage'] = \
                    self.configuration_log['queued_stages'].pop(0)

        self._save_configuration_log()

        # Copy the log to the work directory
        shutil.copy('configuration_log.yaml', self.work_input_path)

    def archive(self):
        """Store model output to laboratory archive and update the
configuration log."""

        # Move files from the restart directory within work to the archive
        # restart directory.
        for f in os.listdir(self.work_restart_path):
            shutil.move(os.path.join(self.work_restart_path, f),
                        self.restart_path)
        os.rmdir(self.work_restart_path)

        # Update the configuration log and save it to the working directory
        completed_stage = self.configuration_log['current_stage']
        self.configuration_log['current_stage'] = ''
        self.configuration_log['completed_stages'].append(completed_stage)

        self._save_configuration_log()

        if len(self.configuration_log["queued_stages"]) == 0:
            # Configuration successfully completed
            os.remove('configuration_log.yaml')

        super(StagedCable, self).archive()

    def collate(self):
        pass

    def _save_configuration_log(self):
        """Write the updated configuration log back to the staging area."""
        with open('configuration_log.yaml', 'w+') as config_log_f:
            yaml.dump(self.configuration_log, config_log_f)



"""payu.experiment
   ===============

   Basic file system operations for Payu

   :copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""

# Standard library
import errno
import os
from pathlib import Path
import re
import shutil
import sys
import shlex
import subprocess
from typing import Union, List
import warnings

# Extensions
import yaml

DEFAULT_CONFIG_FNAME = 'config.yaml'

# Lustre target paths for symbolic paths cannot be 60 characters (yes, really)
# Delete this once this bug in Lustre is fixed
CHECK_LUSTRE_PATH_LEN = True

# File extensions to script interpreters
EXTENSION_TO_INTERPRETER = {'.py': sys.executable,
                            '.sh': '/bin/bash',
                            '.csh': '/bin/tcsh'}


def mkdir_p(path):
    """Create a new directory; ignore if it already exists."""

    try:
        os.makedirs(path)
    except EnvironmentError as exc:
        if exc.errno != errno.EEXIST:
            raise


def movetree(src, dst, symlinks=False):
    """
    Code based on shutil copytree, but non-recursive
    as uses move for contents of src directory
    """
    names = os.listdir(src)
    os.makedirs(dst)
    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        if symlinks and os.path.islink(srcname):
            linkto = os.readlink(srcname)
            os.symlink(linkto, dstname)
        else:
            shutil.move(srcname, dstname)

    shutil.rmtree(src)


class DuplicateKeyWarnLoader(yaml.SafeLoader):
    def construct_mapping(self, node, deep=False):
        """Add warning for duplicate keys in yaml file, as currently
        PyYAML overwrites duplicate keys even though in YAML, keys
        are meant to be unique
        """
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            value = self.construct_object(value_node, deep=deep)
            if key in mapping:
                warnings.warn(
                    "Duplicate key found in config.yaml: "
                    f"key '{key}' with value '{value}'. "
                    f"This overwrites the original value: '{mapping[key]}'"
                )
            mapping[key] = value

        return super().construct_mapping(node, deep)


def read_config(config_fname=None):
    """Parse input configuration file and return a config dict."""

    if not config_fname:
        config_fname = DEFAULT_CONFIG_FNAME

    try:
        with open(config_fname, 'r') as config_file:
            config = yaml.load(config_file, Loader=DuplicateKeyWarnLoader)

        # NOTE: A YAML file with no content returns `None`
        if config is None:
            config = {}
    except IOError as exc:
        if exc.errno == errno.ENOENT:
            print('payu: warning: Configuration file {0} not found!'
                  .format(config_fname))
            config = {}
        else:
            raise

    collate_config = config.pop('collate', {})

    # Transform legacy collate config options
    if type(collate_config) is bool:
        collate_config = {'enable': collate_config}

    collatestr = 'collate_'
    foundkeys = []
    # Cycle through old collate config and convert to newer dict format
    for key in list(config.keys()):
        if key.startswith(collatestr):
            foundkeys.append(key)
            collate_config[key[len(collatestr):]] = config.pop(key)
    if foundkeys:
        print("Use of these keys is deprecated: {}.".format(
                ", ".join(foundkeys)))
        print("Instead use collate dictionary and subkey "
              "without 'collate_' prefix")

    config['collate'] = collate_config

    # Transform legacy archive config options
    archive_config = config.pop('archive', {})
    if type(archive_config) is bool:
        archive_config = {'enable': archive_config}
    config['archive'] = archive_config

    # Transform legacy modules config options
    modules_config = config.pop('modules', {})
    if type(modules_config) is list:
        modules_config = {'load': modules_config}
    config['modules'] = modules_config
     
    # Local "control" path. Must be set here so it can be
    # scanned for storage points
    config["control_path"] = config.get('control',
                                        os.path.dirname(
                                            os.path.abspath(config_fname)))

    return config


def make_symlink(src_path, lnk_path):
    """Safely create a symbolic link to an input field."""

    # Check for Lustre 60-character symbolic link path bug
    if CHECK_LUSTRE_PATH_LEN:
        src_path = patch_lustre_path(src_path)
        lnk_path = patch_lustre_path(lnk_path)

    # os.symlink will happily make a symlink to a non-existent
    # file, but we don't want that behaviour
    # XXX: Do we want to be doing this?
    if not os.path.exists(src_path):
        return

    try:
        os.symlink(src_path, lnk_path)
    except EnvironmentError as exc:
        if exc.errno != errno.EEXIST:
            raise
        elif not os.path.islink(lnk_path):
            # Warn the user, but do not interrupt the job
            print("Warning: Cannot create symbolic link to {p}; a file named "
                  "{f} already exists.".format(p=src_path, f=lnk_path))
        else:
            # Overwrite any existing symbolic link
            if os.path.realpath(lnk_path) != src_path:
                os.remove(lnk_path)
                os.symlink(src_path, lnk_path)


def splitpath(path):
    """Recursively split a filepath into all directories and files."""

    head, tail = os.path.split(path)
    if tail == '':
        return head,
    elif head == '':
        return tail,
    else:
        return splitpath(head) + (tail,)


def patch_lustre_path(f_path):
    """Patch any 60-character pathnames, to avoid a current Lustre bug."""

    if CHECK_LUSTRE_PATH_LEN and len(f_path) == 60:
        if os.path.isabs(f_path):
            f_path = '/.' + f_path
        else:
            f_path = './' + f_path

    return f_path


def check_exe_path(payu_path, pbs_script):
    """Check a payu executable path is locateable """
    if not os.path.isabs(pbs_script):
        pbs_script = os.path.join(payu_path, pbs_script)

    assert os.path.isfile(pbs_script)

    return pbs_script


def is_conda():
    """Return True if python interpreter is in a conda environment"""

    return os.path.exists(os.path.join(sys.prefix, 'conda-meta'))


def parse_ldd_output(ldd_output):
    """Parses the string output from ldd and returns a dictionary of lib filename and fullpath pairs"""
    needed_libs = {}
    for line in ldd_output.split("\n"):
        word_list = line.split()
        if len(word_list) >= 3 and word_list[1] == '=>':
            needed_libs[word_list[0]] = word_list[2]
    return needed_libs


def required_libs(bin_path):
    """
    Runs ldd command and parses the output.
    This function should only be called once per binary
    i.e. Use a singleton pattern in the caller object.
    PARAMETERS:
        string bin_path: full path to the binary
    RETURN:
        dict: {filename-of-lib: fullpath-of-file}
    """
    cmd = 'ldd {0}'.format(bin_path)
    try: 
        ldd_out = subprocess.check_output(shlex.split(cmd)).decode('ascii')
    except:
        print("payu: error running ldd command on exe path: ", bin_path)
        return {}
    return parse_ldd_output(ldd_out)


def list_archive_dirs(archive_path: Union[Path, str],
                      dir_type: str = "output") -> List[str]:
    """Return a sorted list of restart or output directories in archive"""
    naming_pattern = re.compile(fr"^{dir_type}[0-9][0-9][0-9]+$")

    if isinstance(archive_path, str):
        archive_path = Path(archive_path)

    dirs = []
    for path in archive_path.iterdir():
        real_path = path.resolve()
        if real_path.is_dir() and naming_pattern.match(path.name):
            dirs.append(path.name)

    dirs.sort(key=lambda d: int(d.lstrip(dir_type)))
    return dirs


def run_script_command(script_cmd: str, control_path: Path) -> None:
    """Run a user defined script or command.

    Parameters
    ----------
    script_cmd : string
        String of user-script command defined in configuration file
    control_path : Path
        The control directory of the experiment

    Raises
    ------
    RuntimeError
        If there's was an error running the user-script
    """
    try:
        _run_script(script_cmd, control_path)
    except Exception as e:
        error_msg = f"User defined script/command failed to run: {script_cmd}"
        raise RuntimeError(error_msg) from e


def needs_subprocess_shell(command: str) -> bool:
    """Check if command contains shell specific values. For example, file
    redirections, pipes or logical operators.

    Parameters
    ----------
    command: string
        String of command to run in subprocess call

    Returns
    -------
    bool
        Returns True if command requires a subprocess shell, False otherwise
    """
    shell_values = ['>', '<', '|', '&&', '$', '`']
    for value in shell_values:
        if value in command:
            return True
    return False


def _run_script(script_cmd: str, control_path: Path) -> None:
    """Helper recursive function to attempt running a script command.

    Parameters
    ----------
    script_cmd : string
        The script command to attempt to run in subprocess call
    control_path: Path
        The control directory to use for resolving relative filepaths, if file
        is not found
    """
    # First try to interpret the argument as a full command
    try:
        if needs_subprocess_shell(script_cmd):
            subprocess.check_call(script_cmd, shell=True)
        else:
            subprocess.check_call(shlex.split(script_cmd))
        print(script_cmd)

    except FileNotFoundError:
        # Check if script is a file in the control directory
        cmd = control_path / script_cmd
        if cmd.is_file():
            _run_script(str(cmd), control_path)
        else:
            raise

    except PermissionError:
        # Guess the type of interpreter using the file extension
        _, file_ext = os.path.splitext(script_cmd)
        shell_name = EXTENSION_TO_INTERPRETER.get(file_ext, None)
        if shell_name:
            print(
                f'payu: warning: Assuming that {os.path.basename(script_cmd)} '
                f'is a {os.path.basename(shell_name)} script based on the '
                'filename extension.'
            )

            cmd = f'{shell_name} {script_cmd}'
            _run_script(cmd, control_path)
        else:
            raise



import pdb
import pytest
import shlex

import payu
import payu.cli

from .common import cd, make_random_file, get_manifests
from .common import tmpdir, ctrldir, labdir, workdir
from .common import sweep_work, payu_init, payu_setup
from .common import config as config_orig
from .common import write_config
from .common import make_exe, make_inputs, make_restarts, make_all_files

verbose = True

parser = None


def test_generate_parser():

    global parser

    parser = payu.cli.generate_parser()


def test_parse():

    arguments = shlex.split("payu -h")

    with pytest.raises(SystemExit) as parse_error:
        parser.parse_args(arguments[1:])


def test_parse_list():

    arguments = shlex.split("payu list")

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.list_cmd'
    assert len(args) == 0


def test_parse_setup():

    cmd = 'setup'

    arguments = shlex.split('payu {cmd}'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') is None
    assert args.pop('config_path') is None
    assert args.pop('lab_path') is None
    assert args.pop('force_archive') is False
    assert args.pop('reproduce') is False
    assert args.pop('force') is False
    assert args.pop('metadata_off') is False

    assert len(args) == 0

    # Test long options
    arguments = shlex.split('payu {cmd} '
                            '--model mom '
                            '--config path/to/config.yaml '
                            '--laboratory path/to/lab '
                            '--archive '
                            '--force '
                            '--reproduce '
                            '--metadata-off'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('force_archive') is True
    assert args.pop('reproduce') is True
    assert args.pop('force') is True
    assert args.pop('metadata_off') is True

    assert len(args) == 0

    # Test short options
    arguments = shlex.split('payu {cmd} '
                            '-m mom '
                            '-c path/to/config.yaml '
                            '-l path/to/lab '
                            '-f '
                            '-r '
                            '-M'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('force_archive') is False
    assert args.pop('reproduce') is True
    assert args.pop('force') is True
    assert args.pop('metadata_off') is True

    assert len(args) == 0


def test_parse_run():

    cmd = 'run'

    arguments = shlex.split('payu {cmd}'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') is None
    assert args.pop('config_path') is None
    assert args.pop('lab_path') is None
    assert args.pop('reproduce') is False
    assert args.pop('force') is False
    assert args.pop('init_run') is None
    assert args.pop('n_runs') is None
    assert args.pop('force_prune_restarts') is False

    assert len(args) == 0

    # Test long options
    arguments = shlex.split('payu {cmd} '
                            '--model mom '
                            '--config path/to/config.yaml '
                            '--laboratory path/to/lab '
                            '--force '
                            '--initial 99 '
                            '--nruns 999 '
                            '--reproduce '
                            '--force-prune-restarts'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('reproduce') is True
    assert args.pop('force') is True
    assert args.pop('init_run') == '99'
    assert args.pop('n_runs') == '999'
    assert args.pop('force_prune_restarts') is True

    assert len(args) == 0

    # Test short options
    arguments = shlex.split('payu {cmd} '
                            '-m mom '
                            '-c path/to/config.yaml '
                            '-l path/to/lab '
                            '-f '
                            '-i 99 '
                            '-n 999 '
                            '-r '
                            '-F'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('reproduce') is True
    assert args.pop('force') is True
    assert args.pop('init_run') == '99'
    assert args.pop('n_runs') == '999'
    assert args.pop('force_prune_restarts') is True

    assert len(args) == 0


def test_parse_sweep():

    cmd = 'sweep'

    arguments = shlex.split('payu {cmd}'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') is None
    assert args.pop('config_path') is None
    assert args.pop('lab_path') is None
    assert args.pop('hard_sweep') is False
    assert args.pop('metadata_off') is False

    assert len(args) == 0

    # Test long options
    arguments = shlex.split('payu {cmd} '
                            '--model mom '
                            '--config path/to/config.yaml '
                            '--laboratory path/to/lab '
                            '--hard '
                            '--metadata-off'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('hard_sweep') is True
    assert args.pop('metadata_off') is True

    assert len(args) == 0

    # Test short options
    arguments = shlex.split('payu {cmd} '
                            '-m mom '
                            '-c path/to/config.yaml '
                            '-l path/to/lab '
                            '-M'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('hard_sweep') is False
    assert args.pop('metadata_off') is True

    assert len(args) == 0


def test_parse_collate():

    cmd = 'collate'

    arguments = shlex.split('payu {cmd}'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') is None
    assert args.pop('config_path') is None
    assert args.pop('lab_path') is None
    assert args.pop('init_run') is None
    assert args.pop('dir_path') is None

    assert len(args) == 0

    # Test long options
    arguments = shlex.split('payu {cmd} '
                            '--model mom '
                            '--config path/to/config.yaml '
                            '--laboratory path/to/lab '
                            '--initial 99 '
                            '--directory path/to/files '.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('init_run') == '99'
    assert args.pop('dir_path') == 'path/to/files'

    assert len(args) == 0

    # Test short options
    arguments = shlex.split('payu {cmd} '
                            '-m mom '
                            '-c path/to/config.yaml '
                            '-l path/to/lab '
                            '-i 99 '
                            '-d path/to/files '.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'
    assert args.pop('init_run') == '99'
    assert args.pop('dir_path') == 'path/to/files'

    assert len(args) == 0


def test_parse_init():

    cmd = 'init'

    arguments = shlex.split('payu {cmd}'.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') is None
    assert args.pop('config_path') is None
    assert args.pop('lab_path') is None

    assert len(args) == 0

    # Test long options
    arguments = shlex.split('payu {cmd} '
                            '--model mom '
                            '--config path/to/config.yaml '
                            '--laboratory path/to/lab '.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'

    assert len(args) == 0

    # Test short options
    arguments = shlex.split('payu {cmd} '
                            '-m mom '
                            '-c path/to/config.yaml '
                            '-l path/to/lab '.format(cmd=cmd))

    args = vars(parser.parse_args(arguments[1:]))

    run_cmd = args.pop('run_cmd')

    assert run_cmd.__module__ == 'payu.subcommands.{cmd}_cmd'.format(cmd=cmd)

    assert args.pop('model_type') == 'mom'
    assert args.pop('config_path') == 'path/to/config.yaml'
    assert args.pop('lab_path') == 'path/to/lab'

    assert len(args) == 0



# coding: utf-8
"""payu.models.mom6
   ================

   Driver interface to the MOM6 ocean model.

   :copyright: Copyright 2011 Marshall Ward, see AUTHORS for details
   :license: Apache License, Version 2.0, see LICENSE for details
"""

# Standard library
import os

# Extensions
import f90nml

# Local
from payu.models.fms import Fms
from payu.models.mom_mixin import MomMixin


def mom6_add_parameter_files(model):
    """Add parameter files defined in input.nml to model configuration files.
    Broken out of mom6 class so can be used in other models"""
    input_nml = f90nml.read(os.path.join(model.control_path, 'input.nml'))

    input_namelists = ['MOM_input_nml']
    if 'SIS_input_nml' in input_nml:
        input_namelists.append('SIS_input_nml')

    for input in input_namelists:
        input_namelist = input_nml.get(input, {})
        filenames = input_namelist.get('parameter_filename', [])

        if filenames == []:
            print("payu: warning: MOM6: There are no parameter files "
                  f"listed under {input} in input.nml")

        if isinstance(filenames, str):
            model.config_files.append(filenames)
        else:
            model.config_files.extend(filenames)


class Mom6(MomMixin, Fms):
    """Interface to GFDL's MOM6 ocean model."""

    def __init__(self, expt, name, config):

        # FMS initialisation
        super(Mom6, self).__init__(expt, name, config)

        self.model_type = 'mom6'
        self.default_exec = 'MOM6'

        self.config_files = [
            'input.nml',
            'diag_table',
        ]

        self.optional_config_files = [
            'data_table',
            'field_table'
        ]

    def setup(self):
        # FMS initialisation
        super(Mom6, self).setup()

        # Add parameter files to config files
        mom6_add_parameter_files(self)

        # Copy configuration files over to work path
        self.setup_configuration_files()

        self.init_config()

    def init_config(self):
        """Patch input.nml as a new or restart run."""

        input_fpath = os.path.join(self.work_path, 'input.nml')

        input_nml = f90nml.read(input_fpath)

        if ((self.expt.counter == 0 or self.expt.repeat_run) and
                self.prior_restart_path is None):
            input_type = 'n'
        else:
            input_type = 'r'
        input_nml['MOM_input_nml']['input_filename'] = input_type

        if 'SIS_input_nml' in input_nml:
            input_nml['SIS_input_nml']['input_filename'] = input_type

        f90nml.write(input_nml, input_fpath, force=True)



from io import StringIO
import os
from pathlib import Path
import payu.branch
import pytest
import shutil
import stat
import sys

# Submodules
import payu
import payu.fsops
import payu.laboratory
import payu.envmod

from .common import testdir, tmpdir, ctrldir, labdir, workdir, cd
from .common import make_exe, make_inputs, make_restarts, make_all_files


sys.path.insert(1, '../')
verbose = False

tmptwo = testdir / 'tmp2'


def setup_module(module):
    """
    Put any test-wide setup code in here, e.g. creating test files
    """
    if verbose:
        print("setup_module      module:%s" % module.__name__)

    # Should be taken care of by teardown, in case remnants lying around
    try:
        shutil.rmtree(tmpdir)
    except FileNotFoundError:
        pass

    try:
        tmpdir.mkdir()
        labdir.mkdir()
        ctrldir.mkdir()
    except Exception as e:
        print(e)


def teardown_module(module):
    """
    Put any test-wide teardown code in here, e.g. removing test outputs
    """
    if verbose:
        print("teardown_module   module:%s" % module.__name__)

    try:
        shutil.rmtree(tmpdir)
        print('removing tmp')
        shutil.rmtree(tmptwo)
        print('removing tmp2')
    except Exception as e:
        print(e)


def scantree(path):
    """
    Recursively yield DirEntry objects for given directory.
    https://stackoverflow.com/a/33135143/4727812
    """
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            yield from scantree(entry.path)
        else:
            yield entry


def savetree(path):
    """
    Save a directory tree to a dict
    """
    result = {}
    for entry in scantree(path):
        result[entry.name] = (Path(entry.path).relative_to(path),
                              entry.stat().st_size)
    return(result)


# fsops tests
def test_mkdir_p():
    tmp_dir = os.path.join(os.getcwd(), 'tmp_dir')
    payu.fsops.mkdir_p(tmp_dir)

    # Re-create existing directory
    payu.fsops.mkdir_p(tmp_dir)

    # Raise a non-EEXIST error (e.g. EACCES)
    tmp_tmp_dir = os.path.join(tmp_dir, 'more_tmp')
    os.chmod(tmp_dir, stat.S_IRUSR)
    with pytest.raises(OSError):
        payu.fsops.mkdir_p(tmp_tmp_dir)

    # Cleanup
    os.chmod(tmp_dir, stat.S_IWUSR)
    os.rmdir(tmp_dir)


def test_movetree():

    make_all_files()

    treeinfo = savetree(tmpdir)

    tmp_inode = tmpdir.stat().st_ino

    payu.fsops.movetree(tmpdir, tmptwo)

    # Ensure src directory removed
    assert(not tmpdir.exists())

    # Ensure dst directory has new inode number
    assert(tmp_inode != tmptwo.stat().st_ino)

    # Ensure directory tree faithfully moved
    assert(treeinfo == savetree(tmptwo))

    # Move tmp2 back to tmp
    shutil.move(tmptwo, tmpdir)


def test_read_config():
    config_path = os.path.join('test', 'resources', 'config_mom5.yaml')
    config = payu.fsops.read_config(config_path)

    # Test control_path is not set in read_config
    assert('control_path' in config)
    assert(config['control_path'] == os.path.dirname(
                                         os.path.abspath(config_path)))

    # Raise a non-ENOENT error (e.g. EACCES)
    config_tmp = 'config_tmp.yaml'
    config_file = open(config_tmp, 'w')
    os.chmod(config_tmp, 0)

    with pytest.raises(IOError):
        payu.fsops.read_config(config_tmp)

    os.chmod(config_tmp, stat.S_IWUSR | stat.S_IREAD)
    config_file.close()

    config = payu.fsops.read_config(config_tmp)

    assert(config.pop('collate') == {})
    assert(config.pop('control_path') == os.getcwd())
    assert(config.pop('modules') == {})
    assert(config.pop('archive') == {})
    assert(config == {})

    os.remove(config_tmp)


def test_read_config_modules_legacy_option():
    # Test transform legacy modules option
    config_path = os.path.join('test', 'resources', 'config_legacy_modules.yaml')

    config = payu.fsops.read_config(config_path)
    modules_config = config.get('modules', {})

    assert(modules_config.get('load', []) == ['module_1', 'module_2'])
    assert(modules_config.get('use', []) == [])


def test_read_config_modules_option():
    # Test modules with load/use options is unchanged
    config_path = os.path.join('test', 'resources', 'config_modules.yaml')

    config = payu.fsops.read_config(config_path)
    modules_config = config.get('modules', {})

    assert(modules_config.get('load', []) == ['module_1', 'module_2'])
    assert(modules_config.get('use', []) == ['path/to/module/dir/1', 'path/to/module/dir/2'])


def test_make_symlink():
    tmp_path = 'tmp_file'
    tmp_sym = 'tmp_sym'
    tmp_alt_path = 'tmp_alt'
    tmp_dir = 'tmp_dir'

    # Simple symlink test
    tmp = open(tmp_path, 'w')
    payu.fsops.make_symlink(tmp_path, tmp_sym)

    # Override an existing symlink
    tmp_alt = open(tmp_alt_path, 'w')
    payu.fsops.make_symlink(tmp_alt_path, tmp_sym)

    # Try to create symlink when filename already exists
    # TODO: validate stdout
    sys.stdout = StringIO()
    payu.fsops.make_symlink(tmp_path, tmp_alt_path)
    sys.stdout = sys.__stdout__

    # Raise a non-EEXIST signal (EACCESS)
    tmp_dir_sym = os.path.join(tmp_dir, tmp_sym)
    os.mkdir(tmp_dir)
    os.chmod(tmp_dir, 0)
    with pytest.raises(OSError):
        payu.fsops.make_symlink(tmp_path, tmp_dir_sym)

    # Cleanup
    tmp.close()
    tmp_alt.close()

    os.rmdir(tmp_dir)
    os.remove(tmp_sym)
    os.remove(tmp_path)
    os.remove(tmp_alt_path)


def test_splitpath():

    # Absolute path
    paths = payu.fsops.splitpath('/a/b/c')
    assert(paths == ('/', 'a', 'b', 'c'))

    # Relative path
    paths = payu.fsops.splitpath('a/b/c')
    assert(paths == ('a', 'b', 'c'))

    # Single local path
    paths = payu.fsops.splitpath('a')
    assert(paths == ('a',))


def test_default_lab_path():
    # TODO
    pass


def test_lab_new():
    # TODO: validate stdout
    sys.stdout = StringIO()
    lab = payu.laboratory.Laboratory('model')
    sys.stdout = sys.__stdout__


def test_parse_ldd_output():
    ldd_output_path = os.path.join('test', 'resources', 'sample_ldd_output.txt')
    with open(ldd_output_path, 'r') as f:
        ldd_output = f.read()
    required_libs = payu.fsops.parse_ldd_output(ldd_output)
    assert(len(required_libs) == 4)
    assert(required_libs['libmpi.so.40'] == '/apps/openmpi/4.0.2/lib/libmpi.so.40')


def test_lib_update_lib_if_required():
    required_libs_dict = {
        'libmpi.so.40': '/apps/openmpi/4.0.2/lib/libmpi.so.40',
        'libmpi_usempif08_Intel.so.40': '/apps/openmpi/4.0.2/lib/libmpi_usempif08_Intel.so.40'
    }
    result = payu.envmod.lib_update(required_libs_dict, 'libmpi.so')
    assert(result == 'openmpi/4.0.2')


def test_lib_update_if_nci_module_not_required():
    required_libs_dict = {
         'libmpi.so.40': '/$HOME/spack-microarchitectures.git/opt/spack/linux-rocky8-cascadelake/intel-2019.5.281/openmpi-4.1.5-ooyg5wc7sa3tvmcpazqqb44pzip3wbyo/lib/libmpi.so.40', 
         'libmpi_usempif08.so.40': '/$HOME/exe/spack-microarchitectures.git/opt/spack/linux-rocky8-cascadelake/intel-2019.5.281/openmpi-4.1.5-ooyg5wc7sa3tvmcpazqqb44pzip3wbyo/lib/libmpi_usempif08.so.40',
    }
    result = payu.envmod.lib_update(required_libs_dict, 'libmpi.so')
    assert (result == '')


def test_list_archive_dirs():
    # Create archive directories - mix of valid/invalid names
    archive_dirs = [
        'output000', 'output1001', 'output023',
        'output', 'Output001', 'output1',
        'Restart', 'restart2', 'restart',
        'restart102932', 'restart021', 'restart001',
    ]
    tmp_archive = tmpdir / 'test_archive'
    for dir in archive_dirs:
        (tmp_archive / dir).mkdir(parents=True)

    # Add some files
    (tmp_archive / 'restart005').touch()
    (tmp_archive / 'output005').touch()

    # Add a restart symlink
    tmp_archive_2 = tmpdir / 'test_archive_2'
    source_path = tmp_archive_2 / 'restart999'
    source_path.mkdir(parents=True)
    (tmp_archive / 'restart23042').symlink_to(source_path)

    # Test list output dirs and with string archive path
    outputs = payu.fsops.list_archive_dirs(str(tmp_archive), dir_type="output")
    assert outputs == ['output000', 'output023', 'output1001']

    # Test list restarts
    restarts = payu.fsops.list_archive_dirs(tmp_archive, dir_type="restart")
    assert restarts == ['restart001', 'restart021',
                        'restart23042', 'restart102932']

    # Clean up test archive
    shutil.rmtree(tmp_archive)
    shutil.rmtree(tmp_archive_2)


def test_run_userscript_python_script(tmp_path):
    # Create a simple python script
    python_script = tmp_path / 'test_script.py'
    with open(python_script, 'w') as f:
        f.writelines([
            f"with open('{tmp_path}/output.txt', 'w') as f:\n",
            "   f.write('Test Python user script')"
        ])

    # Test run userscript
    payu.fsops.run_script_command('test_script.py', tmp_path)

    # Check script output
    with open((tmp_path / 'output.txt'), 'r') as f:
        assert f.read() == "Test Python user script"


def test_run_userscript_python_script_with_shebang_header(tmp_path):
    # Create a simple python script
    python_script = tmp_path / 'test_script_with_python_shebang.py'
    with open(python_script, 'w') as f:
        f.writelines([
            '#!/usr/bin/env python\n'
            f"with open('{tmp_path}/output.txt', 'w') as f:\n",
            "   f.write('Test Python user script with python shebang header')"
        ])

    # Make file executable
    os.chmod(python_script, 0o775)

    # Test run userscript
    payu.fsops.run_script_command('test_script_with_python_shebang.py',
                                  tmp_path)

    # Check script output
    with open((tmp_path / 'output.txt'), 'r') as f:
        assert f.read() == "Test Python user script with python shebang header"


def test_run_userscript_bash_script(tmp_path):
    # Create a simple bash script
    bash_script = tmp_path / 'test_script.sh'
    with open(bash_script, 'w') as f:
        f.writelines([
            '#!/bin/bash\n',
            'echo -n "Test bash user script" > output.txt'
        ])

    # Test execute script
    with cd(tmp_path):
        payu.fsops.run_script_command('./test_script.sh', tmp_path)

    # Check script output
    with open((tmp_path / 'output.txt'), 'r') as f:
        assert f.read() == "Test bash user script"


def test_userscript_unknown_extension(tmp_path):
    # Create a text file
    text_file = tmp_path / 'test_txt.txt'
    text_file.touch()

    # Test user script raises an error
    with pytest.raises(RuntimeError):
        payu.fsops.run_script_command(str(text_file), tmp_path)


def test_userscript_non_existent_file(tmp_path):
    # Test user script raises an error
    with pytest.raises(RuntimeError):
        payu.fsops.run_script_command('unknown_userscript.sh',
                                      tmp_path)


def test_run_userscript_python_script_error(tmp_path):
    # Create a python script that'll exit with an error
    python_script = tmp_path / 'test_script_error.py'
    with open(python_script, 'w') as f:
        f.write('raise ValueError("Test that script exits with error")')

    # Test userscript raises an error
    with pytest.raises(RuntimeError):
        payu.fsops.run_script_command('test_script_error.py',
                                      tmp_path)


def test_run_userscript_bash_script_error(tmp_path):
    # Create a bash script that'll exit with an error
    bash_script = tmp_path / 'test_script_error.sh'
    with open(bash_script, 'w') as f:
        f.writelines([
            '#!/bin/bash\n',
            'exit 1'
        ])

    # Test userscript raises an error
    with pytest.raises(RuntimeError):
        payu.fsops.run_script_command('test_script_error.sh',
                                      tmp_path)


def test_run_userscript_command(tmp_path):
    # Create a simple command
    cmd = 'echo -n "some_data" > test.txt'

    # Test payu userscript
    with cd(tmp_path):
        payu.fsops.run_script_command(cmd, tmp_path)

    # Check userscript output
    with open((tmp_path / 'test.txt'), 'r') as f:
        assert f.read() == "some_data"


@pytest.mark.parametrize("command, expected", [
    ('echo "Some Data" > test.txt', True),
    ('ls -l | grep "test"', True),
    ('cmd1 && cmd2', True),
    ('cmd1 || cmd2', True),
    ('echo "Data"', False),
    ('ls -l', False),
    ('some_python_script.py', False),
    ('/bin/bash script.sh', False),
    ('echo $PAYU_ENV_VALUE', True),
    ('echo `date`', True),
    ('python ./some/dir/app.py arg_1 arg_2', False)
])
def test_needs_shell(command, expected):
    assert payu.fsops.needs_subprocess_shell(command) == expected


def test_read_config_yaml_duplicate_key():
    """The PyYAML library is used for reading config.yaml, but use ruamel yaml
    is used in when modifying config.yaml as part of payu checkout
    (ruamel is used to preserve comments and multi-line strings).
    This led to bug #441, where pyyaml allowed duplicate keys but
    ruamel.library raises an error
    """
    # Create a yaml file with a duplicate key
    config_content = """
pbs_flags: value1
pbs_flags: value2
"""
    config_path = tmpdir / "config.yaml"
    with config_path.open("w") as file:
        file.write(config_content)

    # Test read config passes without an error but a warning is raised
    warn_msg = "Duplicate key found in config.yaml: key 'pbs_flags' with "
    warn_msg += "value 'value2'. This overwrites the original value: 'value1'"
    with pytest.warns(UserWarning, match=warn_msg):
        payu.fsops.read_config(config_path)

    restart_path = tmpdir / "restarts"

    # Test add restart to config.yaml does not fail with an error, but
    # still raises another warning that duplicates keys will be deleted
    warn_msg = "Removing any subsequent duplicate keys from config.yaml"
    with pytest.warns(UserWarning, match=warn_msg):
        payu.branch.add_restart_to_config(restart_path, config_path)



"""Experiment post-processing - syncing archive to a remote directory

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

# Standard
import getpass
import glob
import os
import shutil
import subprocess


# Local
from payu.fsops import mkdir_p, list_archive_dirs
from payu.metadata import METADATA_FILENAME


class SourcePath():
    """Helper class for building rsync commands - stores attributes
    of source paths to sync.
    Note: Protected paths are paths that shouldn't be removed
    locally if still running an experiment - i.e last output or last
    permanently archived and subsequent restarts
    """
    def __init__(self, path, protected=False, is_log_file=False):
        self.protected = protected
        self.path = path
        self.is_log_file = is_log_file


class SyncToRemoteArchive():
    """Class used for archiving experiment outputs to a remote directory"""

    def __init__(self, expt):
        self.expt = expt
        self.config = self.expt.config.get('sync', {})

        # Ignore the latest output/restart if flagged
        self.ignore_last = os.environ.get('PAYU_SYNC_IGNORE_LAST', False)

        # Use configured url to flag syncing to remote machine
        self.remote_url = self.config.get('url', None)
        self.remote_syncing = self.remote_url is not None

        self.source_paths = []

    def add_outputs_to_sync(self):
        """Add paths of outputs in archive to sync. The last output is
        protected"""
        outputs = list_archive_dirs(archive_path=self.expt.archive_path,
                                    dir_type='output')
        outputs = [os.path.join(self.expt.archive_path, output)
                   for output in outputs]
        if len(outputs) > 0:
            last_output = outputs.pop()
            if not self.ignore_last:
                # Protect the last output
                self.source_paths.append(SourcePath(path=last_output,
                                                    protected=True))
        self.source_paths.extend([SourcePath(path) for path in outputs])

    def add_restarts_to_sync(self):
        """Add paths and protected paths of restarts in archive to sync.
        Last permanently-archived restart and subsequent restarts are
        protected (as local date-based restart pruning uses the last-saved
        restart as a checkpoint for a datetime)"""
        syncing_restarts = self.config.get('restarts', False)
        syncing_all_restarts = os.environ.get('PAYU_SYNC_RESTARTS', False)
        if not (syncing_all_restarts or syncing_restarts):
            return

        # Get sorted list of restarts in archive
        restarts = list_archive_dirs(archive_path=self.expt.archive_path,
                                     dir_type='restart')
        restarts = [os.path.join(self.expt.archive_path, restart)
                    for restart in restarts]
        if restarts == []:
            return

        # Find all restarts that will be 'permanently archived'
        pruned_restarts = self.expt.get_restarts_to_prune(
            ignore_intermediate_restarts=True)
        saved_restarts = [
            restart for restart in restarts
            if os.path.basename(restart) not in pruned_restarts
        ]

        # Sync only permanently saved restarts unless flagged to sync all
        to_sync = saved_restarts if not syncing_all_restarts else restarts

        # Protect last saved restart and any intermediate restarts
        if to_sync and saved_restarts:
            last_saved_index = to_sync.index(saved_restarts[-1])
            paths = to_sync[:last_saved_index]
            protected_paths = to_sync[last_saved_index:]
        else:
            protected_paths, paths = to_sync, []

        if self.ignore_last:
            # Remove the last restart from sync paths
            if protected_paths and protected_paths[-1] == restarts[-1]:
                protected_paths.pop()

        # Add to sync source paths
        self.source_paths.extend([SourcePath(path=path, protected=True)
                                  for path in protected_paths])
        self.source_paths.extend([SourcePath(path) for path in paths])

    def add_extra_source_paths(self):
        """Add additional paths to sync to remote archive"""
        paths = self.config.get('extra_paths', [])
        if isinstance(paths, str):
            paths = [paths]

        for path in paths:
            matching_paths = glob.glob(path)
            # First check if any matching paths exists
            if matching_paths:
                # Add extra paths to protected paths - so they can't be deleted
                self.source_paths.append(SourcePath(path=path, protected=True))
            else:
                print(f"payu: error: No paths matching {path} found. "
                      "Failed to sync path to remote archive")

    def set_destination_path(self):
        "set or create destination path to sync archive to"
        # Remote path to sync output to
        dest_path = self.config.get('path', None)
        if dest_path is None:
            print("There's is no configured path to sync output to. "
                  "In config.yaml, set:\n"
                  "   sync:\n      path: PATH/TO/REMOTE/ARCHIVE\n"
                  "Replace PATH/TO/REMOTE/ARCHIVE with a unique absolute path "
                  "to sync outputs to. Ensure path is unique to avoid "
                  "overwriting exsiting output!")
            raise ValueError("payu: error: Sync path is not defined.")

        if not self.remote_syncing:
            # Create local destination directory if it does not exist
            mkdir_p(dest_path)
        else:
            # Syncing to remote machine
            remote_user = self.config.get('user', None)
            if remote_user is not None:
                dest_path = f'{remote_user}@{self.remote_url}:{dest_path}'
            else:
                dest_path = f'{self.remote_url}:{dest_path}'

        self.destination_path = dest_path

    def set_excludes_flags(self):
        """Add lists of patterns of filepaths to exclude from sync commands"""
        # Get any excludes
        exclude = self.config.get('exclude', [])
        if isinstance(exclude, str):
            exclude = [exclude]

        excludes = ' '.join(['--exclude ' + pattern for pattern in exclude])

        # Default to exclude uncollated files if collation is enabled
        # This can be over-riden using exclude_uncollated config flag
        exclude_uncollated = self.config.get('exclude_uncollated', None)

        if exclude_uncollated is None:
            collate_config = self.expt.config.get('collate', {})
            collating = collate_config.get('enable', True)
            if collating:
                exclude_uncollated = True

        exclude_flag = "--exclude *.nc.*"
        if (exclude_uncollated and exclude_flag not in excludes
                and exclude_flag not in self.config.get('rsync_flags', [])):
            excludes += " --exclude *.nc.*"

        self.excludes = excludes

    def build_cmd(self, source_path):
        """Given a source path to sync, return a rsync command"""
        if source_path.protected:
            # No local delete option for protected paths
            cmd = f'{self.base_rsync_cmd} {self.excludes} '
        elif source_path.is_log_file:
            cmd = f'{self.base_rsync_cmd} {self.remove_files} '
        else:
            cmd = f'{self.base_rsync_cmd} {self.excludes} {self.remove_files} '

        cmd += f'{source_path.path} {self.destination_path}'
        return cmd

    def run_cmd(self, source_path):
        """Given an source path, build and run rsync command"""
        cmd = self.build_cmd(source_path)
        print(cmd)
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError as e:
            print('payu: Error rsyncing archive to remote directory: '
                  f'Failed running command: {cmd}.')
            # TODO: Raise or return?
            return

        if not source_path.protected and self.remove_local_dirs:
            # Only delete real directories; ignore symbolic links
            path = source_path.path
            if os.path.isdir(path) and not os.path.islink(path):
                print(f"Removing {path} from local archive")
                shutil.rmtree(path)

    def git_runlog(self):
        """Add git runlog to remote archive"""
        add_git_runlog = self.config.get("runlog", True)

        if add_git_runlog:
            # Currently runlog is only set up for local remote archive
            if self.remote_syncing:
                print("payu: error: Syncing the git runlog is not implemented "
                      "for syncing to a remote machine")
                return

            control_path = self.expt.control_path
            runlog_path = os.path.join(self.destination_path, "git-runlog")
            if not os.path.exists(runlog_path):
                # Create a bare repository, if it doesn't exist
                try:
                    print("Creating git-runlog bare repository clone"
                          f" at {runlog_path}")
                    cmd = f"git clone --bare {control_path} {runlog_path}"
                    subprocess.check_call(cmd, shell=True)
                except subprocess.CalledProcessError as e:
                    print("payu: error: Failed to create a bare repository. ",
                          f"Error: {e}")
                    return
            else:
                # Update bare gitlog repo
                try:
                    print(f"Updating git-runlog at {runlog_path}")
                    cmd = f"git push {runlog_path}"
                    subprocess.check_call(cmd, shell=True, cwd=control_path)
                except subprocess.CalledProcessError as e:
                    print("payu: error: Failed to push git runlog to bare "
                          f"repository. Error: {e}")

    def run(self):
        """Build and run rsync cmds to remote remote archive """
        # Add outputs and restarts to source paths to sync
        self.add_outputs_to_sync()
        self.add_restarts_to_sync()

        # Add pbs and error logs to paths
        for log_type in ['error_logs', 'pbs_logs']:
            log_path = os.path.join(self.expt.archive_path, log_type)
            if os.path.isdir(log_path):
                self.source_paths.append(SourcePath(path=log_path,
                                                    is_log_file=True))

        # Add metadata path to protected paths, if it exists
        metadata_path = os.path.join(self.expt.archive_path, METADATA_FILENAME)
        if os.path.isfile(metadata_path):
            self.source_paths.append(SourcePath(path=metadata_path,
                                                protected=True))

        # Add any additional paths to protected paths
        self.add_extra_source_paths()

        # Set rsync command components
        self.set_destination_path()
        self.set_excludes_flags()

        # Set base rsync command
        default_flags = '-vrltoD --safe-links'
        rsync_flags = self.config.get('rsync_flags', default_flags)
        self.base_rsync_cmd = f'rsync {rsync_flags}'

        # Set remove local files/dirs options
        remove_files = self.config.get('remove_local_files', False)
        self.remove_files = '--remove-source-files' if remove_files else ''
        self.remove_local_dirs = self.config.get('remove_local_dirs', False)

        # Build and run all rsync commands
        for source_path in self.source_paths:
            self.run_cmd(source_path)

        # Add git runlog to remote archive
        self.git_runlog()



import copy
import os
from pathlib import Path
import pdb
import pytest
import shutil
import yaml

import payu
import payu.models.test

from .common import cd, make_random_file, get_manifests
from .common import tmpdir, ctrldir, labdir, workdir
from .common import sweep_work, payu_init, payu_setup
from .common import config as config_orig
from .common import write_config
from .common import make_exe, make_inputs, make_restarts, make_all_files

verbose = True

config = copy.deepcopy(config_orig)

def make_config_files():
    """
    Create files required for test model
    """

    config_files = payu.models.test.config_files
    for file in config_files:
        make_random_file(ctrldir/file, 29)


def setup_module(module):
    """
    Put any test-wide setup code in here, e.g. creating test files
    """
    if verbose:
        print("setup_module      module:%s" % module.__name__)

    # Should be taken care of by teardown, in case remnants lying around
    try:
        shutil.rmtree(tmpdir)
    except FileNotFoundError:
        pass

    try:
        tmpdir.mkdir()
        labdir.mkdir()
        ctrldir.mkdir()
        make_all_files()
    except Exception as e:
        print(e)

    write_config(config)


def teardown_module(module):
    """
    Put any test-wide teardown code in here, e.g. removing test outputs
    """
    if verbose:
        print("teardown_module   module:%s" % module.__name__)

    try:
        # shutil.rmtree(tmpdir)
        print('removing tmp')
    except Exception as e:
        print(e)

# These are integration tests. They have an undesirable dependence on each
# other. It would be possible to make them independent, but then they'd
# be reproducing previous "tests", like init. So this design is deliberate
# but compromised. It means when running an error in one test can cascade
# and cause other tests to fail.
#
# Unfortunate but there you go.


def test_init():

    # Initialise a payu laboratory
    with cd(ctrldir):
        payu_init(None, None, str(labdir))

    # Check all the correct directories have been created
    for subdir in ['bin', 'input', 'archive', 'codebase']:
        assert((labdir / subdir).is_dir())


def test_setup():

    # Create some input and executable files
    make_inputs()
    make_exe()

    bindir = labdir / 'bin'
    exe = config['exe']

    make_config_files()

    # Run setup
    payu_setup(lab_path=str(labdir))

    assert(workdir.is_symlink())
    assert(workdir.is_dir())
    assert((workdir/exe).resolve() == (bindir/exe).resolve())
    workdirfull = workdir.resolve()

    config_files = payu.models.test.config_files

    for f in config_files + ['config.yaml']:
        assert((workdir/f).is_file())

    for i in range(1, 4):
        assert((workdir/'input_00{i}.bin'.format(i=i)).stat().st_size
               == 1000**2 + i)

    manifests = get_manifests(ctrldir/'manifests')
    for mfpath in manifests:
        assert((ctrldir/'manifests'/mfpath).is_file())

    # Check manifest in work directory is the same as control directory
    assert(manifests == get_manifests(workdir/'manifests'))

    # Sweep workdir and recreate
    sweep_work()

    assert(not workdir.is_dir())
    assert(not workdirfull.is_dir())

    payu_setup(lab_path=str(labdir))

    assert(manifests == get_manifests(workdir/'manifests'))


def test_setup_restartdir():

    restartdir = labdir / 'archive' / 'restarts'

    # Set a restart directory in config
    config['restart'] = str(restartdir)
    write_config(config)

    make_restarts()

    manifests = get_manifests(ctrldir/'manifests')
    payu_setup(lab_path=str(labdir))

    # Manifests should not match, as have added restarts
    assert(not manifests == get_manifests(ctrldir/'manifests'))


def test_exe_reproduce():

    # Set reproduce exe to True
    config['manifest']['reproduce']['exe'] = True
    write_config(config)
    manifests = get_manifests(ctrldir/'manifests')

    # Run setup with unchanged exe but reproduce exe set to True.
    # Should run without error
    payu_setup(lab_path=str(labdir))

    local_exe_path = list(manifests['exe.yaml'].keys())[0]

    assert(manifests == get_manifests(ctrldir/'manifests'))
    assert((ctrldir / local_exe_path).exists())
    assert(Path(manifests['exe.yaml'][local_exe_path]['fullpath']) ==
           (ctrldir / local_exe_path).resolve())

    bindir = labdir / 'bin'
    exe = config['exe']

    # Update the modification time of the executable, should run ok
    (bindir/exe).touch()

    # Run setup with changed exe
    payu_setup(lab_path=str(labdir))

    # Manifests will have changed as fasthash is altered
    assert(not manifests == get_manifests(ctrldir/'manifests'))

    # Reset manifests "truth"
    manifests = get_manifests(ctrldir/'manifests')

    # Recreate fake executable file
    make_exe()

    # Run setup again, which should raise an error due to changed executable
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        # Run setup with unchanged exe but reproduce exe set to True
        payu_setup(lab_path=str(labdir))

        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    # Change reproduce exe back to False
    config['manifest']['reproduce']['exe'] = False
    write_config(config)

    # Run setup with changed exe but reproduce exe set to False
    payu_setup(lab_path=str(labdir))

    # Check manifests have changed as expected
    assert(not manifests == get_manifests(ctrldir/'manifests'))

    # Create a separate executable file
    new_exe = "mock_exe_with_a_different_name"
    make_exe(new_exe)

    # Change exe path and back to exe reproduce true in config.yaml
    config['manifest']['reproduce']['exe'] = True
    config['exe'] = new_exe
    write_config(config)

    # Run setup again, which should raise an error due to changed executable
    with pytest.raises(SystemExit):
        # Run setup with unchanged exe but reproduce exe set to True
        payu_setup(lab_path=str(labdir))


def test_input_reproduce():

    inputdir = labdir / 'input' / config['input']
    inputdir.mkdir(parents=True, exist_ok=True)

    # Set reproduce input to True
    config['manifest']['reproduce']['exe'] = False
    config['manifest']['reproduce']['input'] = True

    config['exe'] = config_orig['exe']
    write_config(config)
    manifests = get_manifests(ctrldir/'manifests')

    # Run setup with unchanged input reproduce input set to True
    # to make sure works with no changes
    payu_setup(lab_path=str(labdir))
    assert(manifests == get_manifests(ctrldir/'manifests'))

    # Delete input directory from config
    input = config['input']
    del(config['input'])
    write_config(config)

    # Expect setup to fail
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        payu_setup(lab_path=str(labdir))

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    # Assert manifests have not changed
    assert(manifests == get_manifests(ctrldir/'manifests'))

    # Add input directory back into config
    config['input'] = input
    write_config(config)

    # Update modification times for input files
    for i in range(1, 4):
        (inputdir/'input_00{i}.bin'.format(i=i)).touch()

    # Run setup, should work as only fasthash will differ, code then
    # checks full hash and updates fasthash if fullhash matches
    payu_setup(lab_path=str(labdir))

    # Manifests should no longer match as fasthashes have been updated
    assert(not manifests == get_manifests(ctrldir/'manifests'))

    # Reset manifest "truth"
    manifests = get_manifests(ctrldir/'manifests')

    # Re-create input files
    make_inputs()

    # Run setup again, which should raise an error due to changed inputs
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        payu_setup(lab_path=str(labdir))

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    # Change reproduce input back to False
    config['manifest']['reproduce']['input'] = False
    write_config(config)

    # Run setup with changed inputs but reproduce input set to False
    payu_setup(lab_path=str(labdir))

    # Check manifests have changed as expected and input files
    # linked in work
    assert(not manifests == get_manifests(ctrldir/'manifests'))
    for i in range(1, 4):
        assert((workdir/'input_00{i}.bin'.format(i=i)).is_file())

    # Reset manifest "truth"
    manifests = get_manifests(ctrldir/'manifests')

    # Delete input manifest
    (ctrldir/'manifests'/'input.yaml').unlink()

    # Remove input from config
    del config['input']
    write_config(config)

    # Setup with no input dir and no manifest. Should work ok
    payu_setup(lab_path=str(labdir))

    # Check there are no linked inputs
    for i in range(1, 4):
        assert(not (workdir/'input_00{i}.bin'.format(i=i)).is_file())

    # Set input path back and recreate input manifest
    config['input'] = input
    write_config(config)
    payu_setup(lab_path=str(labdir))

    # Make a new input file
    (inputdir / 'newInputFile').touch()

    # Set reproduce to true
    config['manifest']['reproduce']['input'] = True
    write_config(config)

    # Run setup again, which should raise an error due to new input file
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        payu_setup(lab_path=str(labdir))

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    # Set reproduce manifests back to False
    config['manifest']['reproduce']['input'] = False
    write_config(config)

    # Check setup runs without an error and new input is linked
    payu_setup(lab_path=str(labdir))
    assert (workdir / 'newInputFile').is_file()


def test_restart_reproduce():

    # Set reproduce restart to True
    config['manifest']['reproduce']['input'] = False
    config['manifest']['reproduce']['restart'] = True
    write_config(config)
    manifests = get_manifests(ctrldir/'manifests')

    # Run setup with unchanged restarts
    payu_setup(lab_path=str(labdir))
    assert(manifests == get_manifests(ctrldir/'manifests'))

    restartdir = labdir / 'archive' / 'restarts'
    # Change modification times on restarts
    for i in range(1, 4):
        (restartdir/'restart_00{i}.bin'.format(i=i)).touch()

    # Run setup with touched restarts, should work with modified
    # manifest
    payu_setup(lab_path=str(labdir))

    # Manifests should have changed
    assert(not manifests == get_manifests(ctrldir/'manifests'))

    # Reset manifest "truth"
    manifests = get_manifests(ctrldir/'manifests')

    # Modify restart files
    make_restarts()

    # Run setup again, which should raise an error due to changed restarts
    with pytest.raises(SystemExit):
        payu_setup(lab_path=str(labdir))

    # Set reproduce restart to False
    config['manifest']['reproduce']['restart'] = False
    write_config(config)

    # Run setup with modified restarts reproduce set to False
    payu_setup(lab_path=str(labdir))

    # Manifests should have changed
    assert(not manifests == get_manifests(ctrldir/'manifests'))

    # Set reproduce restart to True
    config['manifest']['reproduce']['restart'] = True
    write_config(config)

    # Add a new restart file to restart directory
    make_restarts(['restart_005.bin'])

    # Run setup again, which should raise an error due to new restart file
    with pytest.raises(SystemExit):
        payu_setup(lab_path=str(labdir))


def test_all_reproduce():

    # Remove reproduce options from config
    del(config['manifest']['reproduce'])
    write_config(config)

    # Run setup
    payu_setup(lab_path=str(labdir))

    manifests = get_manifests(ctrldir/'manifests')

    make_all_files()

    # Run setup with reproduce=True, which should raise an error as
    # all files changed
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        # Run setup with unchanged exe but reproduce exe set to True
        payu_setup(lab_path=str(labdir), reproduce=True)

    # Run setup
    payu_setup(lab_path=str(labdir))

    # Manifests should have changed
    assert(not manifests == get_manifests(ctrldir/'manifests'))


def test_get_all_previous_fullpaths():

    make_all_files()
    make_config_files()

    # Run setup
    payu_setup(lab_path=str(labdir))

    manifests = get_manifests(ctrldir/'manifests')

    sweep_work()

    with cd(ctrldir):
        manifest = payu.manifest.Manifest(config={}, reproduce=False)
        manifest.load_manifests()
        files = manifest.get_all_previous_fullpaths()

    allfiles = []
    for mf in manifests:
        for f in manifests[mf]:
            allfiles.append(manifests[mf][f]['fullpath'])

    assert(set(files) == set(allfiles))


def test_get_hashes():

    make_all_files()
    make_config_files()

    # Run setup
    payu_setup(lab_path=str(labdir))

    manifests = get_manifests(ctrldir/'manifests')

    sweep_work()

    with cd(ctrldir):
        lab = payu.laboratory.Laboratory(lab_path=str(labdir))
        expt = payu.experiment.Experiment(lab, reproduce=False)
        expt.setup()
        hashes = expt.manifest.manifests['input'].get_hashes('md5')

    allhashes = []
    for f in manifests['input.yaml']:
        allhashes.append(manifests['input.yaml'][f]['hashes']['md5'])

    assert(set(hashes) == set(allhashes))


def test_set_hash():

    # Revert to original config
    config = copy.deepcopy(config_orig)
    write_config(config)

    make_all_files()
    make_config_files()

    # Run setup
    payu_setup(lab_path=str(labdir))

    manifests = get_manifests(ctrldir/'manifests')

    sweep_work()

    # Remove existing manifests. Don't support changing
    # hashes and retaining manifests
    shutil.rmtree(ctrldir/'manifests')

    # Change full hash from md5 to sha256
    config['manifest']['fullhash'] = 'sha256'
    write_config(config)

    # Run setup
    payu_setup(lab_path=str(labdir))

    assert(not manifests == get_manifests(ctrldir/'manifests'))

    manifests = get_manifests(ctrldir/'manifests')

    for mf in manifests:
        for f in manifests[mf]:
            assert(manifests[mf][f]['hashes']['sha256'])
            assert(len(manifests[mf][f]['hashes']['sha256']) == 64)

    sweep_work()

    # Remove existing manifests. Don't support changing
    # hashes and retaining manifests
    shutil.rmtree(ctrldir / 'manifests')

    # Change full hash from md5 to binhash
    config['manifest']['fullhash'] = 'binhash'
    write_config(config)

    # Run setup
    payu_setup(lab_path=str(labdir))

    manifests = get_manifests(ctrldir/'manifests')

    for mf in manifests:
        for f in manifests[mf]:
            assert(list(manifests[mf][f]['hashes'].keys()) == ['binhash'])

def test_hard_sweep():

    # Sweep workdir
    sweep_work(hard_sweep=True)

    # Check all the correct directories have been removed
    assert(not (labdir / 'archive' / 'ctrl').is_dir())
    assert(not (labdir / 'work' / 'ctrl').is_dir())



import argparse
from argparse import Namespace
import copy
import os
from pathlib import Path
import shutil
import sys

import pdb
import pytest
import yaml

import payu

from payu.fsops import read_config
from payu.laboratory import Laboratory
from payu.schedulers import pbs
from payu.schedulers import index as scheduler_index

from .common import cd, make_random_file, get_manifests
from .common import tmpdir, ctrldir, labdir, workdir, payudir
from .common import sweep_work, payu_init, payu_setup
from .common import config as original_config
from .common import write_config
from .common import make_exe, make_inputs, make_restarts
from .common import make_payu_exe, make_all_files

verbose = True

config = copy.deepcopy(original_config)


def setup_module(module):
    """
    Put any test-wide setup code in here, e.g. creating test files
    """
    if verbose:
        print("setup_module      module:%s" % module.__name__)

    # Should be taken care of by teardown, in case remnants lying around
    try:
        shutil.rmtree(tmpdir)
    except FileNotFoundError:
        pass

    try:
        tmpdir.mkdir()
        labdir.mkdir()
        ctrldir.mkdir()
        payudir.mkdir()
        make_payu_exe()
    except Exception as e:
        print(e)

    write_config(config)


def teardown_module(module):
    """
    Put any test-wide teardown code in here, e.g. removing test outputs
    """
    if verbose:
        print("teardown_module   module:%s" % module.__name__)

    try:
        shutil.rmtree(tmpdir)
        print('removing tmp')
    except Exception as e:
        print(e)


def test_encode_mount():

    assert(pbs.encode_mount('/test/a') == 'testa')
    assert(pbs.encode_mount('test/a') == 'testa')
    assert(pbs.encode_mount('test/b') == 'testb')


def test_make_mount_string():

    assert(pbs.make_mount_string('testa', 'x00') == 'testa/x00')


def test_find_mounts():

    paths = ['/f/data/x00/', '/tmp/y11/']
    mounts = ['/f/data', '/tmp']

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', 'tmp/y11']))

    # Only return where a match is found
    mounts = ['/f/data']

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', ]))

    # Test with more mounts than there are paths
    paths = ['/f/data/x00/']
    mounts = ['/f/data', '/tmp']

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', ]))

    # Test with duplicate paths
    paths = ['/f/data/x00/', '/f/data/x00/']
    mounts = ['/f/data', '/tmp']

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', ]))

    # Test with longer path
    paths = ['/f/data/x00/fliberty/gibbet', ]
    mounts = ['/f/data', '/tmp']

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', ]))

    # Test without leading slash
    paths = ['f/data/x00/fliberty/gibbet', ]
    mounts = ['f/data', '/tmp']

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', ]))

    # Test without leading slash
    paths = ['/f/data/x00', ]
    mounts = ['f/data', ]

    assert(pbs.find_mounts(paths, mounts) == set())

    # Test without trailing slash
    paths = ['/f/data/x00', ]
    mounts = ['/f/data', ]

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', ]))

    # Test legacy naming that allows for extra characters at the
    # end of mount path
    paths = ['/f/data1a/x00', ]
    mounts = ['/f/data', ]

    assert(pbs.find_mounts(paths, mounts) == set(['fdata/x00', ]))


def test_run():

    # Use new mechanism to return a scheduler
    sched_name = config.get('scheduler', 'pbs')
    sched_type = scheduler_index[sched_name]
    sched = sched_type()

    # Monkey patch pbs_env_init as we don't have a
    # functioning PBS install in travis
    payu.schedulers.pbs.pbs_env_init = lambda: True

    payu.schedulers.pbs.check_exe_path = lambda x, y: y

    # payu_path = os.path.join(os.environ['PWD'], 'bin')
    payu_path = payudir / 'bin'
    # create new path for payu_path to check a000 picked up as storage
    payu_path = Path('/f/data/a000/some/path')
    pbs_vars = {'PAYU_PATH': str(payu_path)}
    # A pretend python interpreter string
    python_exe = '/f/data/m000/python/bin/python'

    # Test pbs generating a PBS command
    with cd(ctrldir):

        payu_cmd = 'payu-run'

        config['storage'] = {}
        config['storage']['test'] = ['x00']
        config['storage']['/f/data'] = ['x00']

        config['control_path'] = '/f/data/xyz999/experiment'
        config['laboratory'] = '/f/data/c000/blah'
        config['shortpath'] = '/f/data/y00'

        config['modules'] = {}
        config['modules']['use'] = ['/f/data/mm01', '/f/data/mm02/test/modules']

        cmd = sched.submit(payu_cmd, config, pbs_vars, python_exe)

        print(cmd)

        # Create test parser
        parser = argparse.ArgumentParser(description='Test')

        # Add the arguments
        parser.add_argument('-q', type=str, required=True)
        parser.add_argument('-P', type=str, required=True)
        parser.add_argument('-N', type=str, required=True)
        parser.add_argument('-v', metavar='KEY-VALUE',
                            nargs='+', required=True)
        parser.add_argument('-j', type=str, required=True)
        parser.add_argument('-l', metavar='KEY=VALUE',
                            nargs='+', action='append', required=True)
        parser.add_argument('remaining', nargs=argparse.REMAINDER)

        args = parser.parse_args(cmd.split()[1:])

        assert(args.N == config['jobname'])
        assert(args.P == config['project'])
        assert(args.q == config['queue'])

        resources = []
        for resource in args.l:
            resources.extend(resource)

        other_resources = {'wd': True}

        resources_found = {}
        for resource in resources:
            try:
                k, v = resource.split('=')
            except ValueError:
                k, v = (resource, True)
            resources_found[k] = v

        # Check all resources specified in config are correct
        for resource in ['walltime', 'ncpus', 'mem']:
            assert(resources_found[resource] == str(config[resource]))

        assert(resources_found['storage'] ==
               ('fdata/a000+fdata/c000+fdata/m000+fdata/mm01+fdata/mm02+' +
                'fdata/x00+fdata/xyz999+fdata/y00+test/x00'))

        # Check other auto-added resources are present
        for resource in other_resources:
            assert(other_resources[resource] == resources_found[resource])

        env = {}
        for env_var in args.v:
            k, v = env_var.split('=')
            env[k] = v

        assert('PAYU_PATH' in env)
        assert(env['PAYU_PATH'] == str(payu_path))

        assert(args.remaining[-2].endswith('python'))
        assert(args.remaining[-1].endswith(payu_cmd))


def test_tenacity():

    # This should fail and do nothing
    pbs.get_qstat_info()



import os
from pathlib import Path
import shutil

import pytest

from payu.laboratory import Laboratory

from .common import cd
from .common import tmpdir, ctrldir
from .common import write_config


@pytest.fixture(autouse=True)
def setup_and_teardown():
    # Create tmp and control directories
    try:
        tmpdir.mkdir()
        ctrldir.mkdir()
    except Exception as e:
        print(e)

    yield

    # Remove tmp directory
    try:
        shutil.rmtree(tmpdir)
    except Exception as e:
        print(e)


@pytest.mark.parametrize(
    "config, expected_shortpath, expected_user, expected_labname",
    [
        (
            # Test shortpath and model name for laboratory
            {
                'shortpath': '..',
                'model': 'test',
            }, '..', '*', 'test'
        ),
        (
            # Test relative laboratory path
            {
                'shortpath': '/scratch/xx00',
                'model': 'test',
                'laboratory': 'lab'
            }, '/scratch/xx00', '*', 'lab'
        ),
        (
            # Test absolute laboratory path
            {
                'shortpath': '/scratch/xx00',
                'model': 'test',
                'laboratory': '/scratch/aa00/user999/lab'
            }, '/scratch/aa00', 'user999', 'lab'
        ),
        (
            # Test user defined in config.yaml
            {
                'shortpath': '/scratch/xx00',
                'model': 'test',
                'user': 'user123'
            }, '/scratch/xx00', 'user123', 'test'
        ),
    ]
)
def test_laboratory_basepath(config, expected_shortpath, expected_user,
                             expected_labname):
    write_config(config)

    with cd(ctrldir):
        lab = Laboratory(None, None, None)
        basepath = Path(lab.basepath)

        # Check shortpath
        assert basepath.parent.parent.match(expected_shortpath)

        # Check userId - if fixed value
        if expected_user != '*':
            assert basepath.parent.name == expected_user

        # Check laboratory name
        assert basepath.name == expected_labname


@pytest.mark.parametrize(
    "config, expected_project, expected_labname",
    [
        (
            # Test default for project is env variable
            {
                'model': 'test',
            }, 'x00', 'test'
        ),
        (
            # Test project in config is used in default shortpath
            {
                'project': 'aa00',
                'model': 'test',
            }, 'aa00', 'test'
        ),
    ]
)
def test_laboratory_basepath_default_shortpath(config, expected_project,
                                               expected_labname):
    write_config(config)

    # Set a PROJECT env variable to x00 get reproducible paths
    os.environ['PROJECT'] = 'x00'

    with cd(ctrldir):
        lab = Laboratory(None, None, None)

        # Check base of shortpath is used
        assert lab.base in ['/short', '/scratch', '.']
        assert lab.basepath.startswith(lab.base)

        basepath = Path(lab.basepath)

        # Check project
        assert basepath.parent.parent.name == expected_project

        # Check lab name
        assert basepath.name == expected_labname



"""Run the `payu clone` command.

:copyright: Copyright 2018 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

from pathlib import Path

from payu.branch import clone
import payu.subcommands.args as args

title = 'clone'
parameters = {'description': ('A wrapper around git clone. Clones a '
                              'control repository and setup new experiment '
                              'metadata')}

arguments = [args.model, args.config, args.laboratory,
             args.keep_uuid, args.clone_branch,
             args.repository, args.local_directory,
             args.new_branch_name, args.restart_path,
             args.parent_experiment, args.clone_start_point]


def transform_strings_to_path(path_str=None):
    return Path(path_str) if path_str is not None else None


def runcmd(model_type, config_path, lab_path, keep_uuid,
           branch, repository, local_directory, new_branch_name, restart_path,
           parent_experiment, start_point):
    """Execute the command."""
    config_path = transform_strings_to_path(config_path)
    restart_path = transform_strings_to_path(restart_path)
    lab_path = transform_strings_to_path(lab_path)
    directory = transform_strings_to_path(local_directory)

    clone(repository=repository,
          directory=directory,
          branch=branch,
          keep_uuid=keep_uuid,
          model_type=model_type,
          config_path=config_path,
          lab_path=lab_path,
          new_branch_name=new_branch_name,
          restart_path=restart_path,
          parent_experiment=parent_experiment,
          start_point=start_point)


runscript = runcmd



"""Experiment branch support for payu's branch, clone and checkout commands

This may generate new experiment ID, updates, sets any
specified configuration in config.yaml and updates work/archive symlinks

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

import os
import warnings
from pathlib import Path
from typing import Optional
import shutil

from ruamel.yaml import YAML, CommentedMap, constructor
import git

from payu.fsops import read_config, DEFAULT_CONFIG_FNAME, list_archive_dirs
from payu.laboratory import Laboratory
from payu.metadata import Metadata, UUID_FIELD, METADATA_FILENAME
from payu.git_utils import GitRepository, git_clone, PayuBranchError

LAB_WRITE_ACCESS_ERROR = """
Failed to initialise laboratory directories. Skipping creating metadata,
setting up restart configuration, and archive/work symlinks.

To fix, first modify/remove the config.yaml options that determine laboratory
path. Then either run 'payu setup' or rerun checkout command with the current
git branch, e.g. `payu checkout -r  <RESTART_PATH> <CURRENT_BRANCH>`
"""

NO_CONFIG_FOUND_MESSAGE = """No configuration file found on this branch.
Skipping adding new metadata file and creating archive/work symlinks.

To find a branch that has a config file, you can:
    - Display local branches by running:
        payu branch
    - Or display remote branches by running:
        payu branch --remote

To checkout an existing branch, run:
    payu checkout BRANCH_NAME
Where BRANCH_NAME is the name of the branch"""


def check_restart(restart_path: Path,
                  archive_path: Optional[Path] = None) -> Optional[Path]:
    """Checks if restart path exists and whether the archive already
    has pre-existing restarts. Returns a resolved restart path

    Parameters
    ----------
    restart_path: Path
        Absolute, or relative, restart path to start experiment from
    archive_path: Optional[Path], default None
        Experiment archive directory to check for pre-existing restarts files

    Returns
    ----------
    Optional[Path]
        Absolute restart path if a valid path, otherwise None
    """

    # Check for valid path
    if not restart_path.exists():
        warnings.warn((f"Given restart path {restart_path} does not "
                       f"exist. Skipping setting 'restart' in config file"))
        return

    # Resolve to absolute path
    restart_path = restart_path.resolve()

    # Check for pre-existing restarts in archive
    if archive_path and archive_path.exists():
        if len(list_archive_dirs(archive_path, dir_type="restart")) > 0:
            warnings.warn((
                f"Pre-existing restarts found in archive: {archive_path}."
                f"Skipping adding 'restart: {restart_path}' to config file"))
            return

    return restart_path


def add_restart_to_config(restart_path: Path, config_path: Path) -> None:
    """Takes restart path and config path, and add 'restart' flag to the
    config file - which is used to start a run if there isn't a pre-existing
    restart in archive"""

    # Default ruamel yaml preserves comments and multiline strings
    try:
        yaml = YAML()
        config = yaml.load(config_path)
    except constructor.DuplicateKeyError as e:
        # PyYaml which is used to read config allows duplicate keys
        # These will get removed when the configuration file is modified
        warnings.warn(
            "Removing any subsequent duplicate keys from config.yaml"
        )
        yaml = YAML()
        yaml.allow_duplicate_keys = True
        config = yaml.load(config_path)

    # Add in restart path
    config['restart'] = str(restart_path)

    # Write modified lines back to config
    yaml.dump(config, config_path)
    print(f"Added 'restart: {restart_path}' to configuration file:",
          config_path.name)


def get_control_path(config_path: Path) -> Path:
    """Given the config path, return the control path"""
    # Note: Control path is set in read_config
    config = read_config(config_path)
    return Path(config.get('control_path'))


def check_config_path(config_path: Optional[Path] = None) -> Optional[Path]:
    """Checks if configuration file exists"""
    if config_path is None:
        config_path = Path(DEFAULT_CONFIG_FNAME)
        config_path.resolve()

    if not config_path.exists() or not config_path.is_file:
        print(NO_CONFIG_FOUND_MESSAGE)
        raise FileNotFoundError(f"Configuration file {config_path} not found")

    return config_path


def checkout_branch(branch_name: str,
                    is_new_branch: bool = False,
                    is_new_experiment: bool = False,
                    keep_uuid: bool = False,
                    start_point: Optional[str] = None,
                    restart_path: Optional[Path] = None,
                    config_path: Optional[Path] = None,
                    control_path: Optional[Path] = None,
                    model_type: Optional[str] = None,
                    lab_path: Optional[Path] = None,
                    parent_experiment: Optional[str] = None) -> None:
    """Checkout branch, setup metadata and add symlinks

    Parameters
    ----------
    branch_name : str
        Name of branch to checkout/create
    is_new_branch: bool, default False
        Create new branch and mark as new experiment
    is_new_experiment: bool, default False
        Create new uuid for this experiment
    keep_uuid: bool, default False
        Keep UUID unchanged, if it exists - this overrides is_new_experiment
        if there is a pre-existing UUID
    start_point: Optional[str]
        Branch name or commit hash to start new branch from
    restart_path: Optional[Path]
        Restart path to start experiment from
    config_path: Optional[Path]
        Path to configuration file - config.yaml
    control_path: Optional[Path]
        Path to control directory - defaults to current working directory
    model_type: Optional[str]
        Type of model - used for creating a Laboratory
    lab_path: Optional[Path]
        Path to laboratory directory
    parent_experiment: Optional[str]
        Parent experiment UUID to add to generated metadata
    """
    if control_path is None:
        control_path = get_control_path(config_path)

    # Checkout branch
    repo = GitRepository(control_path)
    repo.checkout_branch(branch_name, is_new_branch, start_point)

    # Check config file exists on checked out branch
    config_path = check_config_path(config_path)

    # Initialise Lab
    lab = Laboratory(model_type, config_path, lab_path)
    try:
        lab.initialize()
    except PermissionError:
        print(LAB_WRITE_ACCESS_ERROR)
        raise

    # Initialise metadata
    metadata = Metadata(Path(lab.archive_path),
                        branch=branch_name,
                        config_path=config_path)

    # Setup Metadata
    is_new_experiment = is_new_experiment or is_new_branch
    metadata.setup(keep_uuid=keep_uuid,
                   is_new_experiment=is_new_experiment)

    # Gets valid prior restart path
    prior_restart_path = None
    if restart_path:
        prior_restart_path = check_restart(restart_path=restart_path,
                                           archive_path=metadata.archive_path)

    # Create/update and commit metadata file
    metadata.write_metadata(set_template_values=True,
                            restart_path=prior_restart_path,
                            parent_experiment=parent_experiment)

    # Add restart option to config
    if prior_restart_path:
        add_restart_to_config(prior_restart_path, config_path=config_path)

    # Switch/Remove/Add archive and work symlinks
    experiment = metadata.experiment_name
    switch_symlink(Path(lab.archive_path), control_path, experiment, 'archive')
    switch_symlink(Path(lab.work_path), control_path, experiment, 'work')


def switch_symlink(lab_dir_path: Path, control_path: Path,
                   experiment_name: str, sym_dir: str) -> None:
    """Helper function for removing and switching work and archive
    symlinks in control directory"""
    dir_path = lab_dir_path / experiment_name
    sym_path = control_path / sym_dir

    # Remove symlink if it already exists
    if sym_path.is_symlink():
        previous_path = sym_path.resolve()
        sym_path.unlink()
        print(f"Removed {sym_dir} symlink to {previous_path}")

    # Create symlink, if experiment directory exists in laboratory
    if dir_path.exists():
        sym_path.symlink_to(dir_path)
        print(f"Added {sym_dir} symlink to {dir_path}")


def clone(repository: str,
          directory: Path,
          branch: Optional[str] = None,
          start_point: Optional[str] = None,
          new_branch_name: Optional[str] = None,
          keep_uuid: bool = False,
          model_type: Optional[str] = None,
          config_path: Optional[Path] = None,
          lab_path: Optional[Path] = None,
          restart_path: Optional[Path] = None,
          parent_experiment: Optional[str] = None) -> None:
    """Clone an experiment control repository.

    Parameters:
        repository: str
            Git URL or path to Git repository to clone
        directory: Path
            The control directory where the repository will be cloned
        branch: Optional[str]
            Name of branch to checkout during the git clone
        start_point: Optional[str]
            Branch-name/commit/tag to start new branch from
        new_branch_name: Optional[str]
            Name of new branch to create and checkout
        keep_uuid: bool, default False
            Keep UUID unchanged, if it exists
        config_path: Optional[Path]
            Path to configuration file - config.yaml
        control_path: Optional[Path]
            Path to control directory - defaults to current working directory
        model_type: Optional[str]
            Type of model - used for creating a Laboratory
        lab_path: Optional[Path]
            Path to laboratory directory
        restart_path: Optional[Path]
            Restart path to start experiment from
        parent_experiment: Optional[str]
            Parent experiment UUID to add to generated metadata

    Returns: None

    Note: branch, if defined, can be set to avoid initially checking out the
    default branch during git clone - this is useful for repositories where
    the default branch is not a payu configuration.
    """
    # Resolve directory to an absolute path
    control_path = directory.resolve()

    if control_path.exists():
        raise PayuBranchError(
            f"Directory path `{control_path}` already exists. "
            "Clone to a different path, or cd into the existing directory " +
            "and use `payu checkout` if it is the same git repository"
        )

    # git clone the repository
    repo = git_clone(repository, control_path, branch)

    if restart_path:
        # Check path exists and resolve to an absolute path
        restart_path = check_restart(restart_path)

    owd = os.getcwd()
    try:
        # cd into cloned directory
        os.chdir(control_path)

        # Use checkout wrapper
        if new_branch_name is not None:
            # Create and checkout new branch
            checkout_branch(is_new_branch=True,
                            keep_uuid=keep_uuid,
                            branch_name=new_branch_name,
                            restart_path=restart_path,
                            config_path=config_path,
                            control_path=control_path,
                            model_type=model_type,
                            lab_path=lab_path,
                            parent_experiment=parent_experiment,
                            start_point=start_point)
        else:
            # Checkout branch
            if branch is None:
                branch = repo.get_branch_name()

            checkout_branch(branch_name=branch,
                            config_path=config_path,
                            keep_uuid=keep_uuid,
                            restart_path=restart_path,
                            control_path=control_path,
                            model_type=model_type,
                            lab_path=lab_path,
                            is_new_experiment=True,
                            parent_experiment=parent_experiment)
    except PayuBranchError as e:
        # Remove directory if incomplete checkout
        shutil.rmtree(control_path)
        msg = (
            "Incomplete checkout. To run payu clone again, modify/remove " +
            "the checkout new branch flag: --new-branch/-b, or " +
            "checkout existing branch flag: --branch/-B " +
            f"\n  Checkout error: {e}"
        )
        raise PayuBranchError(msg)
    finally:
        # Change back to original working directory
        os.chdir(owd)

    print(f"To change directory to control directory run:\n  cd {directory}")


def get_branch_metadata(branch: git.Head) -> Optional[CommentedMap]:
    """Return dictionary of branch metadata if it exists, None otherwise"""
    for blob in branch.commit.tree.blobs:
        if blob.name == METADATA_FILENAME:
            # Read file contents
            metadata_content = blob.data_stream.read().decode('utf-8')
            return YAML().load(metadata_content)


def contains_config(branch: git.Head) -> bool:
    """Checks if config file in defined in given branch"""
    contains_config = False
    for blob in branch.commit.tree.blobs:
        if blob.name == DEFAULT_CONFIG_FNAME:
            contains_config = True
    return contains_config


def print_branch_metadata(branch: git.Head, verbose: bool = False):
    """Display given Git branch UUID, or if config.yaml or metadata.yaml does
    not exist.

    Parameters:
        branch: git.Head
            Branch object to parse commit tree.
        verbose: bool, default False
            Display entire metadata files
        remote: bool, default False
            Display remote Git branches

    Returns: None
    """
    # Print branch info
    if not contains_config(branch):
        print(f"    No config file found")
        return

    metadata = get_branch_metadata(branch)

    if metadata is None:
        print("    No metadata file found")
        return

    if verbose:
        # Print all non-null metadata values
        for key, value in metadata.items():
            if value:
                print(f'    {key}: {value}')
    else:
        # Print uuid
        uuid = metadata.get(UUID_FIELD, None)
        if uuid is not None:
            print(f"    {UUID_FIELD}: {uuid}")
        else:
            print(f"    No UUID in metadata file")


def list_branches(config_path: Optional[Path] = None,
                  verbose: bool = False,
                  remote: bool = False):
    """Display local Git branches UUIDs.

    Parameters:
        verbose: bool, default False
            Display entire metadata files
        remote: bool, default False
            Display remote Git branches

    Returns: None"""
    control_path = get_control_path(config_path)
    git_repo = GitRepository(control_path)

    current_branch = git_repo.repo.active_branch
    print(f"* Current Branch: {current_branch.name}")
    print_branch_metadata(current_branch, verbose)

    if remote:
        branches = git_repo.remote_branches_dict()
        label = "Remote Branch"
    else:
        branches = git_repo.local_branches_dict()
        label = "Branch"

    for branch_name, branch in branches.items():
        if branch != current_branch:
            print(f"{label}: {branch_name}")
            print_branch_metadata(branch, verbose)



import datetime
import re

from dateutil.relativedelta import relativedelta
import cftime

NOLEAP, GREGORIAN = range(2)


def int_to_date(date):
    """
    Convert an int of form yyyymmdd to a python date object.
    """

    year = date // 10**4
    month = date % 10**4 // 10**2
    day = date % 10**2

    return datetime.date(year, month, day)


def date_to_int(date):
    return date.year * 10**4 + date.month * 10**2 + date.day


def runtime_from_date(start_date, years, months, days, seconds, caltype):
    """
    Get the number of seconds from start date to start date + date_delta.

    Ignores Feb 29 for caltype == NOLEAP.
    """

    end_date = start_date + relativedelta(
        years=years, months=months, days=days
    )
    runtime = end_date - start_date

    if caltype == NOLEAP:
        runtime -= get_leapdays(start_date, end_date)

    return int(runtime.total_seconds() + seconds)


def date_plus_seconds(init_date, seconds, caltype):
    """
    Get a new_date = date + seconds.

    Ignores Feb 29 for no-leap days.
    """

    end_date = init_date + datetime.timedelta(seconds=seconds)

    if caltype == NOLEAP:
        end_date += get_leapdays(init_date, end_date)
        if end_date.month == 2 and end_date.day == 29:
            end_date += datetime.timedelta(days=1)

    return end_date


def get_leapdays(init_date, final_date):
    """
    Find the number of leap days between arbitrary dates. Returns a
    timedelta object.

    FIXME: calculate this instead of iterating.
    """

    curr_date = init_date
    leap_days = 0

    while curr_date != final_date:
        if curr_date.month == 2 and curr_date.day == 29:
            leap_days += 1

        curr_date += datetime.timedelta(days=1)

    return datetime.timedelta(days=leap_days)


def calculate_leapdays(init_date, final_date):
    """Currently unsupported, it only works for differences in years."""

    leap_days = (final_date.year - 1) // 4 - (init_date.year - 1) // 4
    leap_days -= (final_date.year - 1) // 100 - (init_date.year - 1) // 100
    leap_days += (final_date.year - 1) // 400 - (init_date.year - 1) // 400

    # TODO: Internal date correction (e.g. init_date is 1-March or later)

    return datetime.timedelta(days=leap_days)


# TODO: The caltype logic could be simplified if we switched
# to using just a string as the caltype input. Might require reworking of other
# functions for consistency.
def seconds_between_dates(start_date, end_date, caltype_int):
    """
    Calculate the number of seconds between two datetime objects
    with a specified calender type by using cftime datetime objects
    as intermiaries.

    Parameters
    ----------
    start_date: datetime.date
    end_date: datetime.date
    caltype: Integer, either GREGORIAN or NOLEAP

    Returns
    -------
    seconds: Number of seconds between start_date and end_date.
    """
    # Get the cftime string corresponding to the caltype integer

    # TODO: Is it confusing that GREGORIAN means proleptic gregorian?
    if caltype_int == GREGORIAN:
        calendar_str = "proleptic_gregorian"
    elif caltype_int == NOLEAP:
        calendar_str = "noleap"
    else:
        raise ValueError(f"Unrecognized caltype integer {caltype_int}")

    delta = (date_to_cftime(end_date, calendar_str)
             - date_to_cftime(start_date, calendar_str))

    return int(delta.total_seconds())


def date_to_cftime(date, calendar):
    """
    Convert a datetime.datetime object to a cftime.datetime object which
    has the same year, month, day, hour, minute, second values.

    Parameters
    ----------
    date: datetime.date object
    calendar: string specifying a valid cftime calendar type

    Returns
    -------
    date_cf: cftime.datetime object.
    """
    date_cf = cftime.datetime(
        year=date.year,
        month=date.month,
        day=date.day,
        hour=0,
        minute=0,
        second=0,
        calendar=calendar
    )

    return date_cf


def add_year_start_offset_to_datetime(initial_dt, n):
    """Return a cftime datetime at the start of the year, that is n years
    from the initial datetime"""
    return cftime.datetime(
        year=initial_dt.year + n,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
        calendar=initial_dt.calendar,
    )


def add_month_start_offset_to_datetime(initial_dt, n):
    """Return a cftime datetime of the start of the month, that is n months
    from the initial datetime"""
    years_to_add = (initial_dt.month + n - 1) // 12
    months_to_add = n - years_to_add * 12

    return cftime.datetime(
        year=initial_dt.year + years_to_add,
        month=initial_dt.month + months_to_add,
        day=1,
        hour=0,
        minute=0,
        second=0,
        calendar=initial_dt.calendar,
    )


def add_timedelta_fn(timedelta):
    """Returns a function that takes initial datetime and multiplier n,
    and returns a datetime that is n * offset from the initial datetime"""
    return lambda initial_dt, n: initial_dt + n * timedelta


class DatetimeOffset:
    """A utility class for adding various time offsets to cftime datetimes.

    Parameters:
        unit (str): The unit of the time offset. Supported units are:
            - "YS" for years (start of the year)
            - "MS" for months (start of the month)
            - "W" for weeks
            - "D" for days
            - "H" for hours
            - "T" for minutes
            - "S" for seconds
        magnitude (int): The magnitude of the time offset.

    Methods:
        - `add_to_datetime(initial_dt: cftime.datetime) -> cftime.datetime`:
          Adds the specified time offset to the given cftime datetime and
          returns the resulting datetime.

    Attributes:
        - unit (str): The unit of the time offset.
        - magnitude (int): The magnitude of the time offset.
    """

    def __init__(self, unit, magnitude):
        supported_datetime_offsets = {
            "YS": add_year_start_offset_to_datetime,
            "MS": add_month_start_offset_to_datetime,
            "W": add_timedelta_fn(datetime.timedelta(weeks=1)),
            "D": add_timedelta_fn(datetime.timedelta(days=1)),
            "H": add_timedelta_fn(datetime.timedelta(hours=1)),
            "T": add_timedelta_fn(datetime.timedelta(minutes=1)),
            "S": add_timedelta_fn(datetime.timedelta(seconds=1)),
        }
        if unit not in supported_datetime_offsets:
            raise ValueError(
                f"Unsupported datetime offset: {unit}. "
                "Supported offsets: YS, MS, W, D, H, T, S"
            )
        self.unit = unit
        self.magnitude = magnitude
        self._add_offset_to_datetime = supported_datetime_offsets[unit]

    def add_to_datetime(self, initial_dt):
        """Takes an initial cftime datetime,
        and returns a datetime with the offset added"""

        if not (isinstance(initial_dt, cftime.datetime)):
            raise TypeError(
                f"Invalid initial datetime type: {type(initial_dt)}. "
                "Expected type: cftime.datetime"
            )

        return self._add_offset_to_datetime(
            initial_dt=initial_dt, n=self.magnitude
        )


def parse_date_offset(offset):
    """Parse a given string date offset string and return an DatetimeOffset"""
    match = re.search("[0-9]+", offset)
    if match is None:
        raise ValueError(
            f"No numerical value given for offset: {offset}"
        )
    n = match.group()
    unit = offset.lstrip(n)
    return DatetimeOffset(unit=unit, magnitude=int(n))



# coding: utf-8

# Model type selection
model = {
    'flags': ('--model', '-m'),
    'parameters': {
        'action':   'store',
        'dest':     'model_type',
        'default':  None,
        'help':     'Model type'
    }
}


# Configuration file
config = {
    'flags': ('--config', '-c'),
    'parameters': {
        'action':   'store',
        'dest':     'config_path',
        'default':  None,
        'help':     'Configuration file path',
    }
}


# Initial run counter
initial = {
    'flags': ('--initial', '-i'),
    'parameters': {
        'action':   'store',
        'dest':     'init_run',
        'default':  None,
        'help':     'Starting run counter',
    }
}


# Number of runs.
nruns = {
    'flags': ('--nruns', '-n'),
    'parameters': {
        'action':   'store',
        'dest':     'n_runs',
        'default':  None,
        'help':     'Number of successive experiments ro run',
    }
}


# Laboratory path
laboratory = {
    'flags': ('--laboratory', '--lab', '-l'),
    'parameters': {
        'action':   'store',
        'dest':     'lab_path',
        'default':  None,
        'help':     'The laboratory, this will over-ride the \
                     value given in config.yaml',
    }
}


force_archive = {
    'flags': ('--archive',),
    'parameters': {
        'action':   'store_true',
        'dest':     'force_archive',
        'help':     'Create archive directory during setup',
    }
}


hard_sweep = {
    'flags': ('--hard',),
    'parameters': {
        'action':   'store_true',
        'dest':     'hard_sweep',
        'help':     'Delete archived output',
    }
}

# Explicitly set output_path
dir_path = {
    'flags': ('--directory', '--dir', '-d'),
    'parameters': {
        'action':   'store',
        'dest':     'dir_path',
        'default':  None,
        'help':     'The output directory, this will over-ride the \
                     directory determined from current run number',
    }
}

# Specify a reproducible run
reproduce = {
    'flags': ('--reproduce', '--repro', '-r'),
    'parameters': {
        'action':   'store_true',
        'dest':     'reproduce',
        'default':  False,
        'help':     'Only run if manifests are correct',
    }
}


# Force run to proceed despite existing directories
force = {
    'flags': ('--force', '-f'),
    'parameters': {
        'action':   'store_true',
        'dest':     'force',
        'default':  False,
        'help':     'Force run to proceed, overwriting existing directories',
    }
}

# Force restarts to be pruned despite changes to configuration
force_prune_restarts = {
    'flags': ('--force-prune-restarts', '-F'),
    'parameters': {
        'action':   'store_true',
        'dest':     'force_prune_restarts',
        'default':  False,
        'help':     'Force outdated restart directories to be pruned during \
                    archive, ignoring changes made to configuration.',
    }
}

# Flag for syncing all restarts
sync_restarts = {
    'flags': {'--sync-restarts'},
    'parameters': {
        'action':   'store_true',
        'dest':     'sync_restarts',
        'default':  False,
        'help':     'Sync all restarts in archive to remote directory.',
    }
}

# Flag for ignoring the latest outputs during syncing
sync_ignore_last = {
    'flags': {'--sync-ignore-last'},
    'parameters': {
        'action':   'store_true',
        'dest':     'sync_ignore_last',
        'default':  False,
        'help':     'Ignore the latest outputs and restarts in archive during \
                     syncing.',
    }
}

# Clone Repository
repository = {
    'flags': [],
    'parameters': {
        'dest': 'repository',
        'help': 'The repository to clone from. This can be either a local \
                 path or git url'
    }
}

# Clone to directory
local_directory = {
    'flags': [],
    'parameters': {
        'dest': 'local_directory',
        'help': 'The directory to clone into'
    }
}

# Clone uuid flag
keep_uuid = {
    'flags': ('-k', '--keep-uuid'),
    'parameters': {
        'action':   'store_true',
        'default':  False,
        'dest': 'keep_uuid',
        'help': 'If an experiment uuid exists, leave it unchanged'
    }
}

# Clone branch
clone_branch = {
    'flags': ('--branch', '-B'),
    'parameters': {
        'action':   'store',
        'dest':     'branch',
        'default':  None,
        'help': 'Clone and checkout this branch'
    }
}

# Clone branch
clone_start_point = {
    'flags': ('--start-point', '-s'),
    'parameters': {
        'action':   'store',
        'dest': 'start_point',
        'default':  None,
        'help': 'New branch will start from this commit or tag'
    }
}

# Clone create branch
new_branch_name = {
    'flags': ('--new-branch', '-b'),
    'parameters': {
        'action':   'store',
        'dest':  'new_branch_name',
        'default': None,
        'help': 'The name of the git branch to create and checkout'
    }
}

# Parent experiment UUID
parent_experiment = {
    'flags': ('--parent-experiment', '-p'),
    'parameters': {
        'action':   'store',
        'dest':  'parent_experiment',
        'default': None,
        'help': 'The parent experiment UUID to add to generated metadata'
    }
}

# Branch name
branch_name = {
    'flags': [],
    'parameters': {
        'dest': 'branch_name',
        'help': 'The name of the git branch to create/checkout'
    }
}

# Branch start point
start_point = {
    'flags': [],
    'parameters': {
        'nargs': '?',
        'dest': 'start_point',
        'help': 'The new branch head will point to this commit'
    }
}


# Branch starting restart
restart_path = {
    'flags': ('--restart', '-r'),
    'parameters': {
        'dest': 'restart_path',
        'action': 'store',
        'help': 'The restart path from which to start the model run'
    }
}

# Checkout new branch flag
new_branch = {
    'flags': ['-b'],
    'parameters': {
        'dest': 'new_branch',
        'action': 'store_true',
        'default':  False,
        'help': 'Create new branch'
    }
}

# List branches verbose flag
verbose = {
    'flags': ['--verbose', '-v'],
    'parameters': {
        'dest': 'verbose',
        'action': 'store_true',
        'default':  False,
        'help': 'Display all contents of metadata file'
    }
}

# List remote branches flag
remote = {
    'flags': ['--remote', '-r'],
    'parameters': {
        'dest': 'remote',
        'action': 'store_true',
        'default':  False,
        'help': 'Display metadata of branches in remote directory'
    }
}


# Disable metadata + UUID generation
metadata_off = {
    'flags': ['--metadata-off', '-M'],
    'parameters': {
        'dest': 'metadata_off',
        'action': 'store_true',
        'default': False,
        'help': 'Disable experiment metadata and UUID generation and commits'
    }
}


""" Functions to support PBS based schedulers

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

# Standard library
import os
import re
import sys
import shlex
import subprocess

import payu.envmod as envmod
from payu.fsops import check_exe_path
from payu.manifest import Manifest
from payu.schedulers.scheduler import Scheduler

from tenacity import retry, stop_after_delay


# TODO: This is a stub acting as a minimal port to a Scheduler class.
class PBS(Scheduler):
    # TODO: __init__

    def submit(self, pbs_script, pbs_config, pbs_vars=None, python_exe=None):
        """Prepare a correct PBS command string"""

        pbs_env_init()

        # Initialisation
        if pbs_vars is None:
            pbs_vars = {}

        # Necessary for testing
        if python_exe is None:
            python_exe = sys.executable

        pbs_flags = []

        pbs_queue = pbs_config.get('queue', 'normal')
        pbs_flags.append('-q {queue}'.format(queue=pbs_queue))

        pbs_project = pbs_config.get('project', os.environ['PROJECT'])
        pbs_flags.append('-P {project}'.format(project=pbs_project))

        pbs_resources = ['walltime', 'ncpus', 'mem', 'jobfs']

        for res_key in pbs_resources:
            res_flags = []
            res_val = pbs_config.get(res_key)
            if res_val:
                res_flags.append(
                    '{key}={val}'.format(key=res_key, val=res_val)
                )
            if res_flags:
                pbs_flags.append('-l {res}'.format(res=','.join(res_flags)))

        # TODO: Need to pass lab.config_path somehow...
        pbs_jobname = pbs_config.get('jobname', os.path.basename(os.getcwd()))
        if pbs_jobname:
            # PBSPro has a 15-character jobname limit
            pbs_flags.append('-N {name}'.format(name=pbs_jobname[:15]))

        pbs_priority = pbs_config.get('priority')
        if pbs_priority:
            pbs_flags.append('-p {priority}'.format(priority=pbs_priority))

        pbs_flags.append('-l wd')

        pbs_join = pbs_config.get('join', 'n')
        if pbs_join not in ('oe', 'eo', 'n'):
            print('payu: error: unknown qsub IO stream join setting.')
            sys.exit(-1)
        else:
            pbs_flags.append('-j {join}'.format(join=pbs_join))

        # Append environment variables to qsub command
        # TODO: Support full export of environment variables: `qsub -V`
        pbs_vstring = ','.join('{0}={1}'.format(k, v)
                               for k, v in pbs_vars.items())
        pbs_flags.append('-v ' + pbs_vstring)

        storages = set()
        storage_config = pbs_config.get('storage', {})
        mounts = set(['/scratch', '/g/data'])
        for mount in storage_config:
            mounts.add(mount)
            for project in storage_config[mount]:
                storages.add(make_mount_string(encode_mount(mount), project))

        # Append any additional qsub flags here
        pbs_flags_extend = pbs_config.get('qsub_flags')
        if pbs_flags_extend:
            pbs_flags.append(pbs_flags_extend)

        payu_path = pbs_vars.get('PAYU_PATH', os.path.dirname(sys.argv[0]))
        pbs_script = check_exe_path(payu_path, pbs_script)
        ctrl_path = pbs_config.get('control_path')

        # Check for storage paths that might need to be mounted in the
        # python and script paths
        extra_search_paths = [python_exe, payu_path, pbs_script, ctrl_path]

        laboratory_path = pbs_config.get('laboratory', None)
        if laboratory_path is not None:
            extra_search_paths.append(laboratory_path)
        short_path = pbs_config.get('shortpath', None)
        if short_path is not None:
            extra_search_paths.append(short_path)

        module_use_paths = pbs_config.get('modules', {}).get('use', [])
        extra_search_paths.extend(module_use_paths)

        remote_sync_directory = pbs_config.get('sync', {}).get('path', None)
        if remote_sync_directory is not None:
            extra_search_paths.append(remote_sync_directory)
        storages.update(find_mounts(extra_search_paths, mounts))
        storages.update(find_mounts(get_manifest_paths(), mounts))

        # Add storage flags. Note that these are sorted to get predictable
        # behaviour for testing
        pbs_flags_extend = '+'.join(sorted(storages))
        if pbs_flags_extend:
            pbs_flags.append("-l storage={}".format(pbs_flags_extend))

        # Set up environment modules here for PBS.
        envmod.setup()
        envmod.module('load', 'pbs')

        # Construct job submission command
        cmd = 'qsub {flags} -- {python} {script}'.format(
            flags=' '.join(pbs_flags),
            python=python_exe,
            script=pbs_script
        )

        return cmd


# TODO: These support functions can probably be integrated into the class

def get_job_id(short=True):
    """
    Return PBS job id
    """

    jobid = os.environ.get('PBS_JOBID', '')

    if short:
        # Strip off '.rman2'
        jobid = jobid.split('.')[0]

    return(jobid)


def get_job_info():
    """
    Get information about the job from the PBS server
    """
    jobid = get_job_id()

    info = None

    if not jobid == '':
        info = get_qstat_info('-ft {0}'.format(jobid), 'Job Id:')

    if info is not None:
        # Select the dict for this job (there should only be one
        # entry in any case)
        info = info['Job Id: {}'.format(jobid)]

        # Add the jobid to the dict and then return
        info['Job_ID'] = jobid

    return info


def pbs_env_init():

    # Initialise against PBS_CONF_FILE
    if sys.platform == 'win32':
        pbs_conf_fpath = r'C:\Program Files\PBS Pro\pbs.conf'
    else:
        pbs_conf_fpath = '/etc/pbs.conf'
    os.environ['PBS_CONF_FILE'] = pbs_conf_fpath

    try:
        with open(pbs_conf_fpath) as pbs_conf:
            for line in pbs_conf:
                try:
                    key, value = line.split('=')
                    os.environ[key] = value.rstrip()
                except ValueError:
                    pass
    except IOError as ec:
        print('Unable to find PBS_CONF_FILE ... ' + pbs_conf_fpath)
        sys.exit(1)


# Wrap this in retry from tenancity. Keep trying for 10 seconds and
# even if still fails return None
@retry(stop=stop_after_delay(10), retry_error_callback=lambda a: None)
def get_qstat_info(qflag, header, projects=None, users=None):

    qstat = os.path.join(os.environ['PBS_EXEC'], 'bin', 'qstat')
    cmd = '{} {}'.format(qstat, qflag)

    cmd = shlex.split(cmd)
    output = subprocess.check_output(cmd)
    if sys.version_info.major >= 3:
        output = output.decode()

    entries = (e for e in output.split('{}: '.format(header)) if e)

    # Immediately remove any non-project entries
    if projects or users:
        entries = (e for e in entries
                   if any('project = {}'.format(p) in e for p in projects)
                   or any('Job_Owner = {}'.format(u) in e for u in users))

    attribs = ((k.split('.')[0], v.replace('\n\t', '').split('\n'))
               for k, v in (e.split('\n', 1) for e in entries))

    status = {k: dict((kk.strip(), vv.strip())
              for kk, vv in (att.split('=', 1) for att in v if att))
              for k, v in attribs}

    return status


def encode_mount(mount):
    """
    Turn a mount point point into the keyword used to specify storages,
    i.e. remove path separators
    """
    return re.sub(os.path.sep, '', mount)


def make_mount_string(mount, project):
    """
    Return mount and project string used to specify storages
    """
    return "{mount}/{project}".format(mount=mount, project=project)


def find_mounts(paths, mounts):
    """
    Search a path for a matching mount point and return a set of unique
    NCI compatible strings to add to the qsub command
    """
    if not isinstance(paths, list):
        paths = [paths, ]
    if not isinstance(mounts, set):
        mounts = set(mounts)

    storages = set()

    for p in paths:
        for m in mounts:
            if p.startswith(m):
                # Find the number of path elements in the mount string
                offset = len(m.split(os.path.sep))
                # Relevant project code is the next element of the path
                # after the mount point. DO NOT USE os.path.split as it
                # is not consistent with trailing slash
                proj = p.split(os.path.sep)[offset]
                storages.add(make_mount_string(encode_mount(m), proj))
                break

    return storages


def get_manifest_paths():
    """
    Return a list of paths from manifest files to use to check for
    storage paths
    """
    tmpmanifest = Manifest(config={}, reproduce=False)
    tmpmanifest.load_manifests()

    return tmpmanifest.get_all_previous_fullpaths()
